# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

__all__=['packageManager','display','services','utils']

import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")

    try:
        #Check if we have an python shell available, if not, use our ProxyShell
        get_ipython()
    except NameError:
        from .proxyShell import ProxyInteractiveShell
        ProxyInteractiveShell.instance()   

    #shortcut to logging
    import pixiedust.utils.pdLogging as pdLogging
    logger = pdLogging.getPixiedustLogger()
    getLogger = pdLogging.getLogger

    from pixiedust.utils.environment import Environment
    if Environment.hasSpark:
        #shortcut to packageManager
        import pixiedust.packageManager as packageManager
        printAllPackages=packageManager.printAllPackages
        installPackage=packageManager.installPackage
        uninstallPackage=packageManager.uninstallPackage

        try:
            from py4j.protocol import Py4JJavaError
            #javaBridge and scalaBridge only work in the driver, not an executor
            from pixiedust.utils.javaBridge import *
            from pixiedust.utils.scalaBridge import *

            #shortcut to Spark job monitoring
            from pixiedust.utils.sparkJobProgressMonitor import enableSparkJobProgressMonitor
            enableJobMonitor = enableSparkJobProgressMonitor
        except (NameError, Py4JJavaError):
            #IPython not available we must be in a spark executor
            pass

    #automated import into the user namespace
    try:
        from IPython.core.getipython import get_ipython
        import pixiedust.display
        import pixiedust.services
        get_ipython().user_ns["display"]=display.display

        from pixiedust.utils.sampleData import sampleData
        from pixiedust.utils import checkVersion
        checkVersion()
    except (NameError):
        #IPython not available we must be in a spark executor
        pass



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.utils.template import PixiedustTemplateEnvironment
from IPython.core.getipython import *
from IPython.display import display, HTML, Javascript
from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.utils.environment import Environment
from functools import reduce
import uuid
import json
import sys
import traceback
import pixiedust
from IPython.core.getipython import get_ipython
from collections import OrderedDict
from threading import Thread, Lock, Event
import time

myLogger = pixiedust.getLogger(__name__)
_env = PixiedustTemplateEnvironment()
progressMonitor = None
loadingProgressMonitor = False

def enableSparkJobProgressMonitor():
    if Environment.isRunningOnDSX:
        print("Spark Job Progress Monitoring cannot be started on DSX")
        return

    global progressMonitor, loadingProgressMonitor
    if progressMonitor is None and not loadingProgressMonitor:
        loadingProgressMonitor = True
        def startSparkJobProgressMonitor():
            global progressMonitor
            progressMonitor = SparkJobProgressMonitor()
        t = Thread(target=startSparkJobProgressMonitor)
        t.daemon = True
        t.start()
        print("Succesfully enabled Spark Job Progress Monitor")
    else:
        print("Spark Job Progress Monitor already enabled")

class SparkJobProgressMonitorOutput(Thread):
    class Java:
        implements = ["com.ibm.pixiedust.PixiedustOutputListener"]
    
    def __init__(self):
        super(SparkJobProgressMonitorOutput,self).__init__()
        self.prefix = None
        self.lock = Lock()
        self.triggerEvent = Event()
        self.daemon = True
        self.progressData = OrderedDict()

    def getUpdaterId(self):
        return "updaterId{0}".format(self.prefix)

    def getProgressHTMLId(self):
        return "progress{0}".format(self.prefix)

    def run(self):
        while True:
            self.triggerEvent.wait()
            with self.lock:
                self.triggerEvent.clear()
                if bool(self.progressData):
                    progressData = self.progressData
                    self.progressData = OrderedDict()
                else:
                    progressData = OrderedDict()

            if bool(progressData):
                js = ""
                for data in progressData.values():
                    channel = data["channel"]
                    if channel=="jobStart":
                        js += _env.getTemplate("sparkJobProgressMonitor/addJobTab.js").render( 
                            prefix=self.prefix, data=data, overalNumTasks=reduce(lambda x,y:x+y["numTasks"], data["stageInfos"], 0) 
                        )
                    elif channel=="stageSubmitted":
                        js += _env.getTemplate("sparkJobProgressMonitor/updateStageStatus.js").render( 
                            prefix=self.prefix, stageId=data["stageInfo"]["stageId"], status="Submitted", host=None 
                        )
                    elif channel=="taskStart":
                        js += _env.getTemplate("sparkJobProgressMonitor/taskStart.js").render( prefix=self.prefix, data=data, increment = data["increment"] )
                        js += "\n"
                        js += _env.getTemplate("sparkJobProgressMonitor/updateStageStatus.js").render( 
                            prefix=self.prefix, stageId=data["stageId"], status="Running",
                            host="{0}({1})".format(data["taskInfo"]["executorId"],data["taskInfo"]["host"] )
                        )
                    elif channel=="stageCompleted":
                        js += _env.getTemplate("sparkJobProgressMonitor/updateStageStatus.js").render( 
                            prefix=self.prefix, stageId=data["stageInfo"]["stageId"], status="Completed", host=None 
                        )
                    elif channel=="jobEnd":
                        js += _env.getTemplate("sparkJobProgressMonitor/jobEnded.js").render( 
                            prefix=self.prefix, jobId=data["jobId"] 
                        )
                    js += "\n"

                display(Javascript(js))
            time.sleep(0.5)

    def display_with_id(self, obj, display_id, update=False):
        """Create a new display with an id"""
        ip = get_ipython()
        if hasattr(ip, "kernel"):
            data, md = ip.display_formatter.format(obj)
            content = {
                'data': data,
                'metadata': md,
                'transient': {'display_id': display_id},
            }
            msg_type = 'update_display_data' if update else 'display_data'
            ip.kernel.session.send(ip.kernel.iopub_socket, msg_type, content, parent=ip.parent_header)
        else:
            display(obj)

    def printOutput(self, s):
        print(s)

    def sendChannel(self, channel, data):
        self.printStuff(channel, data)

    def onRunCell(self):
        self.prefix = str(uuid.uuid4())[:8]
        #Create the place holder area for the progress monitor
        self.display_with_id( 
            HTML( _env.getTemplate("sparkJobProgressMonitor/pmLayout.html").render( prefix = self.prefix)),self.getProgressHTMLId() 
        )

    def printStuff(self,channel, s):
        try:
            data = json.loads(s)
            data["channel"] = channel
            data["increment"] = 1
            key = None
            if channel=="jobStart":
                key = "{0}-{1}".format(channel,data["jobId"])
            elif channel=="stageSubmitted":
                key = "{0}-{1}".format(channel,data["stageInfo"]["stageId"])
            elif channel=="taskStart":
                key = "{0}-{1}".format(channel,data["stageId"])
            elif channel=="stageCompleted":
                key = "{0}-{1}".format(channel,data["stageInfo"]["stageId"])
            elif channel=="jobEnd":
                key = "{0}-{1}".format(channel,data["jobId"])

            if key:
                with self.lock:
                    if key in self.progressData:
                        data["increment"] = self.progressData[key]["increment"] + 1
                    self.progressData[key] = data
                    self.triggerEvent.set()
        except:
            print("Unexpected error: {0} - {1} : {2}".format(channel, s, sys.exc_info()[0]))
            traceback.print_exc()

class SparkJobProgressMonitor(object):
    def __init__(self):
        self.monitorOutput = None
        self.addSparkListener()
        self.displayRuns={}
        self.newDisplayRun = False

    def onDisplayRun(self, contextId):
        if contextId is None or self.monitorOutput is None:
            self.newDisplayRun=True
            return

        cellContext = self.displayRuns.get( contextId )
        if cellContext and cellContext != self.monitorOutput.prefix:
            #switch the cell context if not a new display Run
            if self.newDisplayRun:
                self.displayRuns.pop( contextId, None )
            else:
                self.monitorOutput.prefix = cellContext
        elif cellContext is None:
            self.displayRuns[contextId] = self.monitorOutput.prefix

        if cellContext:
            display(Javascript(_env.getTemplate("sparkJobProgressMonitor/emptyTabs.js").render(prefix=cellContext)))

        self.newDisplayRun=False

    def addSparkListener(self):
        try:
            get_ipython().run_cell_magic(
                "scala",
                "cl=sparkProgressMonitor noSqlContext",
                _env.getTemplate("sparkJobProgressMonitor/addSparkListener.scala").render()
            )

            listener = get_ipython().user_ns.get("__pixiedustSparkListener")

            #access the listener object from the namespace
            if listener:
                self.monitorOutput = SparkJobProgressMonitorOutput()
                self.monitorOutput.start()
                #Add pre_run_cell event handler
                get_ipython().events.register('pre_run_cell',lambda: self.monitorOutput.onRunCell() )
                listener.setChannelListener( self.monitorOutput )
        except:
            myLogger.exception("Unexpected error while adding Spark Listener")
            raise



from .connectionWidget import ConnectionWidget
from pixiedust.display.app import *
from pixiedust.services.serviceManager import *
from pyspark import *
from pyspark.sql import *
from xml.sax.saxutils import escape
import base64
import json
import requests

view_dbs = '0'
view_db = '1'
view_db_all_docs = '2'
view_db_all_docs_page = '3'
view_db_design_docs = '4'
view_db_query = '5'
view_db_query_results = '6'
view_db_query_results_page = '7'
view_db_search = '8'
view_db_search_results = '9'
view_db_search_results_page = '10'
view_db_view = '11'
view_db_view_page = '12'
view_generate_dataframe_all_docs = '13'
view_generate_dataframe_query = '14'
view_generate_dataframe_search = '15'
view_generate_dataframe_view = '16'


@PixieApp
class CloudantBrowser(ConnectionWidget):

    def get_data_frame(self):
        return self.df

    @route()
    def default(self):
        return """<div pd_widget="DataSourcesList"/>"""

    @route(selectedConnection='*')
    def start(self):
        connection = getConnection('cloudant', self.selectedConnection)
        credentials = json.loads(connection['PAYLOAD'])['credentials']
        self.spark_session = None
        self.sql_context = None
        self.host = credentials['host']
        self.protocol = credentials.get('protocol', 'https')
        self.port = credentials.get('port', 443)
        self.username = credentials['username']
        self.password = credentials['password']
        self.selectedConnection = None
        return self._view_dbs()

    @route(view=view_dbs)
    def _view_dbs(self):
        output = """
<div class="row">
    <div class="col-sm-10" style="padding: 10px;">
        <p><a href="#" pd_options="view=None">Back</a></p>
        <h3>Databases</h3>
"""
        self.all_docs_limit = 5
        self.all_docs_skip = 0
        self.query = None
        self.query_limit = 5
        self.query_skip = 0
        self.search_query = None
        self.search_limit = 5
        self.search_max_limit = 200
        self.search_skip = 0
        self.search_bookmark = None
        self.search_prev_bookmark = None
        self.view_name = None
        self.view_limit = 5
        self.view_skip = 0
        dbs = self.get_all_dbs(self.host, self.username, self.password)
        for db in dbs:
            output += '<p>'
            output += '<h4 style="margin: 0px"><a href="#" pd_options="view={}">{}'.format(view_db, db)
            output += '<pd_script>self.db="{}"</pd_script>'.format(db)
            output += '</a></h4>'
            output += '</p>'
        output += """
    </div>
</div>
"""
        return output

    @route(view=view_db)
    def _view_db(self):
        search_index_html = ''
        view_list_html = ''
        response = self.get_design_docs(self.host, self.username, self.password, self.db, -1)
        if 'rows' in response.keys():
            all_design_docs = map(lambda x: x['doc'], response['rows'])
            search_index_html = self.get_db_search_index_list_html(all_design_docs)
            view_list_html = self.get_db_view_list_html(all_design_docs)
        return """
<div class="row">
    <div class="col-sm-2" style="padding: 10px;">
        <p><a href="#" pd_options="view=""" + view_dbs + """">Back</a></p>
        <b><h3>""" + self.db + """</h3></b>
        <p>
            <h4 style="margin: 0px">
                <a href="#">All Documents
                    <target pd_target="target{{prefix}}" pd_options="view=""" + view_db_all_docs + """" />
                </a>              
            </h4>
        </p>
         <p>
            <h4 style="margin: 0px">
                <a href="#">Query
                    <target pd_target="target{{prefix}}" pd_options="view=""" + view_db_query + """" />
                </a>
            </h4>
        </p>
        <p>
            <h4 style="margin: 0px">
                <a href="#">Design Documents
                    <target pd_target="target{{prefix}}" pd_options="view=""" + view_db_design_docs + """" />
                </a>
            </h4>
        </p>
""" + search_index_html + """
""" + view_list_html + """
    </div>
    <div id="target{{prefix}}" class="col-sm-10" style="padding: 0px 10px 0px 30px;">
    </div>
</div>
"""

    def get_db_search_index_list_html(self, all_design_docs):
        return self.get_db_ddoc_list_html(all_design_docs, 'Search Indexes', 'indexes', 'search_doc', 'search_index', view_db_search)

    def get_db_view_list_html(self, all_design_docs):
        return self.get_db_ddoc_list_html(all_design_docs, 'Views', 'views', 'view_doc', 'view_name', view_db_view)

    def get_db_ddoc_list_html(self, all_design_docs, title, design_doc_type, doc_var_name, type_var_name, route_name):
        html = ''
        docs = []
        for doc in all_design_docs:
            if design_doc_type in doc.keys():
                docs.append(doc)
        if len(docs) > 0:
            html += """
<p><h4 style="margin: 0px">""" + title + """</h3></p>
"""
            i = -1
            for doc in docs:
                i = i + 1
                html += """
<p>
    <a href="#""" + design_doc_type + str(i) + """-{{prefix}}" data-toggle="collapse" style="text-decoration: none;"><h4 style="margin: 0px"><b>""" + doc['_id'] + """</b></h4></a>
    <div id=\"""" + design_doc_type + str(i) + """-{{prefix}}" class="collapse in" style="padding-top: 10px;">
"""
                for key in doc[design_doc_type]:
                    html += """
        <p>
            <h4 style="margin-top: 0px; margin-left: 20px;">
                <a href="#">""" + key + """
                    <target pd_target="target{{prefix}}" pd_options="view=""" + route_name + """" />
                    <pd_script>self.""" + doc_var_name + """='""" + doc['_id'][len('_design/'):] + """'
self.""" + type_var_name + """='""" + key + """'</pd_script>
                </a>
            </h4>
        </p>
"""
            html += """
    </div>
</p>"""
        return html

    @route(view=view_db_all_docs)
    def _view_all_docs(self):
        self.all_docs_skip = 0
        output = self.get_dialog_chrome_top('All Documents', None)
        output += """
<div class="row">
    <div id="target2{{prefix}}">""" + self._view_all_docs_page() + """</div>
</div>
"""
        output += self.get_dialog_chrome_bottom()
        return output

    @route(view=view_db_all_docs_page)
    def _view_all_docs_page(self):
        response = self.get_all_docs(self.host, self.username, self.password, self.db, self.all_docs_limit, self.all_docs_skip)
        if 'rows' in response.keys() and len(response['rows']) > 0:
            output = self.format_docs(map(lambda row: row['doc'], response['rows']))
            # previous/next buttons
            prev_skip = self.all_docs_skip
            next_skip = self.all_docs_skip
            if 'total_rows' in response.keys() and response['total_rows'] > (self.all_docs_skip + self.all_docs_limit):
                next_skip = next_skip + self.all_docs_limit
            if prev_skip > 0:
                prev_skip = prev_skip - self.all_docs_limit
            show_next_button = (next_skip > self.all_docs_skip)
            show_prev_button = (prev_skip < self.all_docs_skip)
            if show_prev_button or show_next_button:
                output += """
<div style="margin-top: 10px;">"""
                if show_prev_button:
                    output +="""
    <button type="submit" class="btn btn-info">Previous
        <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_all_docs_page + """" />
        <pd_script>self.all_docs_skip=""" + str(prev_skip) + """</pd_script>
    </button>"""
                if show_next_button:
                    output +="""
    <button type="submit" class="btn btn-info">Next
        <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_all_docs_page + """" />
        <pd_script>self.all_docs_skip=""" + str(next_skip) + """</pd_script>
    </button>"""
                output += '<span style="padding-left: 5px"><b>{} to {} of {}</b></span>'.format(self.all_docs_skip+1, self.all_docs_skip+len(response['rows']), response['total_rows'])
                output += """
</div>"""
            output += self.get_dataframe_button(view_generate_dataframe_all_docs, 'target2')
        else:
            output = """<p><b>No documents found.</b></p>"""
            output += self.get_dataframe_button(None, None)
        return output

    @route(view=view_db_design_docs)
    def _view_design_docs(self):
        output = self.get_dialog_chrome_top('Design Documents', None)
        output += """
<div class="row">
    <div id="target2{{prefix}}">"""
        response = self.get_design_docs(self.host, self.username, self.password, self.db, -1)
        if 'rows' in response.keys():
            output += self.format_docs(map(lambda row: row['doc'], response['rows']))
        else:
            output = """<p><b>No documents found.</b></p>"""
        output += """
    </div>
</div>"""
        output += self.get_dialog_chrome_bottom()
        return output

    @route(view=view_db_query)
    def _view_query(self):
        if self.query is None:
            self.query ="""
{
  "selector": {"_id": {"$gt": 0}},
  "fields": ["_id", "_rev"],
  "sort": [{"_id": "asc"}]
}"""
        self.query_skip = 0
        output = self.get_dialog_chrome_top('Query', None)
        output += """
<div class="row">
    <div class="form-group">
        <textarea rows="6" class="form-control" id="query{{prefix}}" style="font-family: monospace;">""" + self.query + """</textarea>
    </div>
    <div class="form-group">  
        <button type="submit" class="btn btn-primary">Go
            <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_query_results + """" />
            <pd_script>self.query="$val(query{{prefix}})"</pd_script>
        </button>
    </div>
</div>
<div class="row">
    <div id="target2{{prefix}}"></div>
</div>
"""
        output += self.get_dialog_chrome_bottom()
        return output

    @route(view=view_db_query_results)
    def _view_query_results(self):
        self.query_skip = 0
        return self._view_query_results_page()

    @route(view=view_db_query_results_page)
    def _view_query_results_page(self):
        self.query = self.query.replace('&quot;', '"')
        response = self.run_query(self.host, self.username, self.password, self.db,
                                  self.query, self.query_limit, self.query_skip)
        if 'docs' in response.keys() and len(response['docs']) > 0:
            output = self.format_docs(response['docs'])
            # previous/next buttons
            prev_skip = self.query_skip
            next_skip = self.query_skip
            if len(response['docs']) >= self.query_limit:
                next_skip = next_skip + self.query_limit
            if prev_skip > 0:
                prev_skip = prev_skip - self.query_limit
            show_next_button = (next_skip > self.query_skip)
            show_prev_button = (prev_skip < self.query_skip)
            if show_prev_button or show_next_button:
                output += """
<div style="margin-top: 10px;">"""
                if show_prev_button:
                    output +="""
    <button type="submit" class="btn btn-info">Previous
        <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_query_results_page + """" />
        <pd_script>self.query_skip=""" + str(prev_skip) + """</pd_script>
    </button>"""
                if show_next_button:
                    output +="""
    <button type="submit" class="btn btn-info">Next
        <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_query_results_page + """" />
        <pd_script>self.query_skip=""" + str(next_skip) + """</pd_script>
    </button>"""
                output += '<span style="padding-left: 5px"><b>{} to {}</b></span>'.format(self.query_skip+1, self.query_skip+len(response['docs']))
                output += """
</div>"""
            output += self.get_dataframe_button(view_generate_dataframe_query, 'target2')
        else:
            output = """<b>No documents found.</b>"""
            output += self.get_dataframe_button(None, None)
        return output

    @route(view=view_db_search)
    def _view_search(self):
        if self.search_query is None:
            self.search_query = ''
        self.search_bookmark = None
        self.search_prev_bookmark=None
        output = self.get_dialog_chrome_top('Search', '{}/{}'.format(self.search_doc, self.search_index))
        output += """
<div class="row">
    <div class="form-group">
        <input type="text" class="form-control" id="search{{prefix}}" value="{{self.search_query}}" />
    </div>
    <div class="form-group">
        <button type="submit" class="btn btn-primary">Go
            <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_search_results + """" />
            <pd_script>self.search_query="$val(search{{prefix}})"</pd_script>
        </button>
    </div>
</div>
<div class="row">
    <div id="target2{{prefix}}"></div>
</div>
"""
        output += self.get_dialog_chrome_bottom()
        return output

    @route(view=view_db_search_results)
    def _view_search_results(self):
        self.search_bookmark = None
        self.search_skip = 0
        return self._view_search_results_page()

    @route(view=view_db_search_results_page)
    def _view_search_results_page(self):
        response = self.run_search(self.host, self.username, self.password, self.db,
                                   self.search_doc, self.search_index, self.search_query,
                                   self.search_limit, self.search_bookmark)
        if 'rows' in response.keys() and len(response['rows']) > 0:
            output = self.format_docs(map(lambda x: x['doc'], response['rows']))
            # previous/next buttons
            if 'bookmark' in response.keys() and len(response['rows']) >= self.search_limit:
                next_bookmark = response['bookmark']
            else:
                next_bookmark = None
            if self.search_bookmark is not None or next_bookmark is not None:
                output += """
<div style="margin-top: 10px;">"""
                if self.search_bookmark is not None:
                    if self.search_prev_bookmark is None:
                        prev_bookmark_str = 'None'
                    else:
                        prev_bookmark_str = '"{}"'.format(self.search_prev_bookmark)
                    output +="""
    <button type="submit" class="btn btn-info">Previous
        <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_search_results_page + """" />
        <pd_script>self.search_bookmark=""" + prev_bookmark_str + """
self.search_skip=""" + str(self.search_skip - self.search_limit) + """</pd_script>
    </button>"""
                if next_bookmark is not None:
                    if self.search_bookmark is None:
                        prev_bookmark_str = 'None'
                    else:
                        prev_bookmark_str = '"{}"'.format(self.search_bookmark)
                    if next_bookmark is None:
                        next_bookmark_str = 'None'
                    else:
                        next_bookmark_str = '"{}"'.format(next_bookmark)
                    output +="""
    <button type="submit" class="btn btn-info">Next
        <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_search_results_page + """" />
        <pd_script>self.search_prev_bookmark=""" + prev_bookmark_str + """
self.search_bookmark=""" + next_bookmark_str + """
self.search_skip=""" + str(self.search_skip + self.search_limit) + """</pd_script>
    </button>"""
                output += '<span style="padding-left: 5px"><b>{} to {}</b></span>'.format(self.search_skip+1, self.search_skip+len(response['rows']))
                output += """
</div>"""
            output += self.get_dataframe_button(view_generate_dataframe_search, 'target2')
        else:
            output = """<b>No documents found.</b>"""
            output += self.get_dataframe_button(None, None)
        return output

    @route(view=view_db_view)
    def _view_view(self):
        self.view_skip = 0
        output = self.get_dialog_chrome_top('View', '{}/{}'.format(self.view_doc, self.view_name))
        output += """
<div class="row">
    <div id="target2{{prefix}}">""" + self._view_view_page() + """</div>
</div>
"""
        output += self.get_dialog_chrome_bottom()
        return output

        return self._view_view_page()

    @route(view=view_db_view_page)
    def _view_view_page(self):
        response = self.get_view(self.host, self.username, self.password, self.db,
                                 self.view_doc, self.view_name, self.view_limit, self.view_skip)
        if 'rows' in response.keys():
            output = self.format_docs(map(lambda x: x['doc'], response['rows']))
            # previous/next buttons
            prev_skip = self.view_skip
            next_skip = self.view_skip
            if 'total_rows' in response.keys() and response['total_rows'] > (self.view_skip + self.view_limit):
                next_skip = next_skip + self.view_limit
            if prev_skip > 0:
                prev_skip = prev_skip - self.view_limit
            show_next_button = (next_skip > self.view_skip)
            show_prev_button = (prev_skip < self.view_skip)
            if show_prev_button or show_next_button:
                output += """
<div style="margin-top: 10px;">"""
                if show_prev_button:
                    output +="""
    <button type="submit" class="btn btn-info">Previous
        <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_view_page + """" />
        <pd_script>self.view_skip=""" + str(prev_skip) + """</pd_script>
    </button>"""
                if show_next_button:
                    output +="""
    <button type="submit" class="btn btn-info">Next
        <target pd_target="target2{{prefix}}" pd_options="view=""" + view_db_view_page + """" />
        <pd_script>self.view_skip=""" + str(next_skip) + """</pd_script>
    </button>"""
                output += '<span style="padding-left: 5px"><b>{} to {} of {}</b></span>'.format(self.view_skip+1, self.view_skip+len(response['rows']), response['total_rows'])
                output += """
</div>"""
            output += self.get_dataframe_button(view_generate_dataframe_view, 'target2')
        else:
            output = """<p><b>No documents found.</b></p>"""
            output += self.get_dataframe_button(None, None)
        output += self.get_dialog_chrome_bottom()
        return output

    @route(view=view_generate_dataframe_all_docs)
    def _generate_dataframe_all_docs(self):
        self.df = self.get_dataframe_reader().load(self.db)
        output = 'DataFrame generated. Access by calling app.get_data_frame()'.format(self.db)
        return """
<p><b>""" + output + """</b></p>
"""

    @route(view=view_generate_dataframe_query)
    def _generate_dataframe_query(self):
        response = self.run_query(self.host, self.username, self.password, self.db, self.query, -1, 0)
        if 'docs' in response.keys():
            self.df = self.get_dataframe_builder().createDataFrame(response['docs'])
            output = 'DataFrame generated. Access by calling app.get_data_frame()'.format(self.db)
        else:
            output = 'No documents found.'
        return """
<p><b>""" + output + """</b></p>
"""

    @route(view=view_generate_dataframe_search)
    def _generate_dataframe_search(self):
        docs = []
        bookmark = None
        while True:
            response = self.run_search(self.host, self.username, self.password, self.db,
                                       self.search_doc, self.search_index, self.search_query,
                                       self.search_max_limit, bookmark)
            if 'rows' in response.keys():
                if len(response['rows']) == 0:
                    break
                docs.extend(map(lambda x: x['doc'], response['rows']))
                if 'bookmark' in response.keys():
                    bookmark = response['bookmark']
                else:
                    break
            else:
                break
        if len(docs) > 0:
            self.df = self.get_dataframe_builder().createDataFrame(docs)
            output = 'DataFrame generated. Access by calling app.get_data_frame()'.format(self.db)
        else:
            output = 'No documents found.'
        return """
<p><b>""" + output + """</b></p>"""

    @route(view=view_generate_dataframe_view)
    def _generate_dataframe_view(self):
        self.df = self.get_dataframe_reader() \
            .option("index", "_design/{}/_view/{}".format(self.view_doc, self.view_name)) \
            .load(self.db)
        output = 'DataFrame generated. Access by calling app.get_data_frame()' \
            .format(self.db, self.view_doc, self.view_name)
        return """
<p><b>""" + output + """</b></p>
"""

    def get_all_dbs(self, host, username, password):
        url = '{0}://{1}/_all_dbs'.format(self.protocol, host)
        r = requests.get(url, headers={
            'Authorization': 'Basic {}'.format(base64.b64encode('{}:{}'.format(username, password))),
            'Accept': 'application/json'
        })
        return r.json()

    def get_all_docs(self, host, username, password, db, limit, skip):
        url = '{}://{}/{}/_all_docs?include_docs=true'.format(self.protocol, host,db)
        if limit > 0:
            url = '{}&limit={}'.format(url,limit)
        if skip > 0:
            url = '{}&skip={}'.format(url,skip)
        r = requests.get(url, headers={
            'Authorization': 'Basic {}'.format(base64.b64encode('{}:{}'.format(username, password))),
            'Accept': 'application/json'
        })
        return r.json()

    def get_design_docs(self, host, username, password, db, limit):
        url = '{}://{}/{}/_all_docs?startkey="_design/"&endkey="_design0"&include_docs=true'.format(self.protocol, host,db)
        if limit > 0:
            url = '{}&limit={}'.format(url,limit)
        r = requests.get(url, headers={
            'Authorization': 'Basic {}'.format(base64.b64encode('{}:{}'.format(username, password))),
            'Accept': 'application/json'
        })
        return r.json()

    def run_query(self, host, username, password, db, query, limit, skip):
        if limit > 0:
            query_json = json.loads(query)
            query_json['limit'] = limit
            if skip > 0:
                query_json['skip'] = skip
            query = json.dumps(query_json)
        url = '{}://{}/{}/_find'.format(self.protocol, host, db)
        r = requests.post(url, data=query, headers={
            'Authorization': 'Basic {}'.format(base64.b64encode('{}:{}'.format(username, password))),
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        })
        return r.json()

    def run_search(self, host, username, password, db, design_doc, index, query, limit, bookmark):
        url = '{}://{}/{}/_design/{}/_search/{}?q={}&include_docs=true'.format(self.protocol, host, db, design_doc, index, query)
        if limit > 0:
            url = '{}&limit={}'.format(url,limit)
            if bookmark:
                url = '{}&bookmark={}'.format(url,bookmark)
        r = requests.get(url, headers={
            'Authorization': 'Basic {}'.format(base64.b64encode('{}:{}'.format(username, password))),
            'Accept': 'application/json'
        })
        return r.json()

    def get_view(self, host, username, password, db, design_doc, view, limit, skip):
        url = '{}://{}/{}/_design/{}/_view/{}?include_docs=true'.format(self.protocol, host, db, design_doc, view)
        if limit > 0:
            url = '{}&limit={}'.format(url, limit)
        if skip > 0:
            url = '{}&skip={}'.format(url, skip)
        r = requests.get(url, headers={
            'Authorization': 'Basic {}'.format(base64.b64encode('{}:{}'.format(username, password))),
            'Accept': 'application/json'
        })
        return r.json()

    def format_docs(self, docs):
        output = ''
        i = -1
        for doc in docs:
            i = i + 1
            j = 0
            content = json.dumps(doc, indent=2, sort_keys=True).encode('utf-8')
            content = self.escape_html(content)
            title = '<pre style="font-size: 9pt;">' + doc['_id'] + ' {</pre>'
            for key in sorted(doc.keys()):
                if key == '_id' or key == '_rev' or isinstance(doc[key], list):
                    continue
                else:
                    try:
                        value = str(doc[key].encode('utf-8'))
                        if len(value) > 20:
                            value = value[0:20] + '...'
                        title += '<pre style="font-size: 9pt;">  "{}": "{}"</pre>'.format(key, value)
                        j = j + 1
                        if j >= 3:
                            break
                    except:
                        pass
            title += '<pre>   ...</pre>'
            output += '<div style="margin-top: 5px;">'
            output += '<a href="#doc' + str(i) + '-{{prefix}}" style="text-decoration: none;" data-toggle="collapse">' + title + '</a>'
            output += '<div id="doc' + str(i) + '-{{prefix}}" class="collapse">'
            output += '<pre style="margin: 0px 20px 0px 20px; padding: 10px; font-size: 9pt; background-color: white; max-height: 200px; white-space: pre; overflow: scroll; border-style: solid; border-radius: 2px; border-width: 1px; border-color: #aaa;">{}</pre>'.format(content)
            output += '</div>'
            output += '<a href="#doc' + str(i) + '-{{prefix}}" style="text-decoration: none;" data-toggle="collapse"><pre style="font-size: 9pt;">}</pre></a>'
            output += '</div>'
        return output

    def get_dialog_chrome_top(self, title, subtitle):
        subtitle_str = ''
        if subtitle:
            subtitle_str = ' ({})'.format(subtitle)
        return """
<div class="row" style="padding: 0x; margin: 10px; border-style: solid; border-radius: 4px 4px 0 0; border-width: 1px; border-color: #ddd;">
    <h4 style="padding: 0px 0px 10px 10px;">""" + title + subtitle_str + """</h4>
    <div id="target{{prefix}}" class="row" style="padding: 15px; margin-bottom: 0px; background-color: #f5f5f5;">"""

    def get_dialog_chrome_bottom(self):
        return """
    </div>
</div>"""
        return output

    def get_dataframe_button(self, generate_dataframe_view, generate_dataframe_target):
        output = ''
        if generate_dataframe_view is not None:
            output += """
<hr style="border: 1px solid #ddd;">
<button type="submit" class="btn btn-primary" data-dismiss="modal">Generate DataFrame
    <target pd_target=\"""" + generate_dataframe_target + """{{prefix}}" pd_options="view=""" + generate_dataframe_view + """" />
</button>"""
        return output

    def escape_html(self, text):
        return escape(text)

    def get_dataframe_reader(self):
        if self.sql_context is None and self.spark_session is None:
            self.get_dataframe_builder()
        if self.spark_session is not None:
            return self.spark_session.read.format("com.cloudant.spark")
        else:
            return self.sql_context.read.format("com.cloudant.spark") \
                .option("cloudant.host", self.host) \
                .option("cloudant.username", self.username) \
                .option("cloudant.password", self.password)

    def get_dataframe_builder(self):
        if self.sql_context is None and self.spark_session is None:
            spark_context = SparkContext.getOrCreate()
            spark_major_version = int(spark_context.version[0:spark_context.version.index('.')])
            if spark_major_version >= 2:
                self.spark_session = SparkSession.builder \
                    .config("cloudant.host", self.host) \
                    .config("cloudant.username", self.username) \
                    .config("cloudant.password", self.password) \
                    .getOrCreate()
            else:
                self.sql_context = SQLContext(spark_context)
        if self.spark_session is not None:
            return self.spark_session
        else:
            return self.sql_context



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Inherited from maven-artifact https://github.com/hamnis/maven-artifact
# -------------------------------------------------------------------------------

import hashlib
import os
from .package import Package
from lxml import etree
import uuid
from IPython.display import display, HTML, Javascript

try:
    from urllib.request import Request, urlopen, URLError, HTTPError
except ImportError:
    from urllib2 import Request, urlopen, URLError, HTTPError

def doRequest(url, onFail, onSuccess, username=None, password=None):
        headers = {"User-Agent": "PixieDust PackageManager/1.0"}
        if username and password:
            import base64
            headers["Authorization"] = "Basic " + base64.b64encode(self.username + ":" + self.password)
        req = Request(url, None, headers)
        try:
            response = urlopen(req)
        except HTTPError as e:
            onFail(url, e)
        except URLError as e:
            onFail(url, e)
        else:
            return onSuccess(response)

class Resolver(object):
    def __init__(self, base):
        if base.endswith("/"):
            base = base.rstrip("/")
        self.base = base

    def _parseMetadataXML(self, r):
        try:
            return etree.parse(r)
        except e:
            print(e)

    def _find_latest_version_available(self, package):
        path = "/%s/maven-metadata.xml" % (package.path(False))
        xml = doRequest(self.base + path, self._onFail, lambda r: self._parseMetadataXML(r))
        v = xml.xpath("/metadata/versioning/versions/version[last()]/text()")
        if v:
            return v[0]

    def _onFail(self, url, e):
        raise RequestException("Failed to download maven-metadata.xml from '%s'" % url)

    def resolve(self, package):
        version = package.version
        if not package.version or package.version == "latest":            
            version = self._find_latest_version_available(package)
        return Package.clone(package)

    def uri_for_artifact(self, package):        
        resolved = self.resolve(package)
        return package.getUri(self.base, resolved.version)

class RequestException(Exception):
    def __init__(self, msg):
        self.msg = msg

class Downloader(object):
    def __init__(self, base="http://repo1.maven.org/maven2", username=None, password=None):
        self.username = username
        self.password = password
        self.resolver = Resolver(base)
        self.prefix = str(uuid.uuid4())[:8]
    
    def download(self, package, filename=None, suppress_log=False):
        filename = package.getFilePath(filename)
        url = self.resolver.uri_for_artifact(package)
        if not self.verify_md5(filename, url + ".md5"):
            if not suppress_log:
                hook=self._chunk_report
            else:
                hook=self._chunk_report_suppress

            onError = lambda uri, err: self._throwDownloadFailed("Failed to download package " + str(package) + "from " + uri)
            response = doRequest(url, onError, lambda r: r)
            
            if response:
                if not suppress_log:
                    print("Downloading package {0} to {1}".format(package, filename))
                with open(filename, 'wb') as f:
                    self._write_chunks(response, f, report_hook=hook)
                return (package, True)
            else:
                return (package, False)
        else:
            if not suppress_log:
                print("%s is already up to date" % package)
            return (package, True)

    def _throwDownloadFailed(self, msg):
        raise RequestException(msg)

    def _chunk_report_suppress(self, bytes_so_far, chunk_size, total_size):
        pass

    def _chunk_report(self, bytes_so_far, chunk_size, total_size):
        if bytes_so_far == 0:
            display( HTML( """
                <div>
                    <span id="pm_label{0}">Starting download...</span>
                    <progress id="pm_progress{0}" max="100" value="0" style="width:200px"></progress>
                </div>""".format(self.prefix)
                )
            )
        else:
            percent = float(bytes_so_far) / total_size
            percent = round(percent*100, 2)
            display(
                Javascript("""
                    $("#pm_label{prefix}").text("{label}");
                    $("#pm_progress{prefix}").attr("value", {percent});
                """.format(prefix=self.prefix, label="Downloaded {0} of {1} bytes".format(bytes_so_far, total_size), percent=percent))
            )

    def _write_chunks(self, response, file, chunk_size=8192, report_hook=None):
        total_size = response.headers['Content-Length'].strip()
        total_size = int(total_size)
        bytes_so_far = 0

        if report_hook:
                report_hook(bytes_so_far, chunk_size, total_size)

        while 1:
            chunk = response.read(chunk_size)
            bytes_so_far += len(chunk)

            if not chunk:
                break

            file.write(chunk)
            if report_hook:
                report_hook(bytes_so_far, chunk_size, total_size)

        return bytes_so_far

    def verify_md5(self, file, remote_md5):
        if not os.path.exists(file):
            return False
        else:
            local_md5 = self._local_md5(file)
            onError = lambda uri, err: _throwDownloadFailed("Failed to download MD5 from " + uri)
            remote = doRequest(remote_md5, onError, lambda r: r.read())
            return local_md5 == remote

    def _local_md5(self, file):
        md5 = hashlib.md5()
        with open(file, 'rb') as f:
            for chunk in iter(lambda: f.read(8192), ''):
                md5.update(chunk)
        return md5.hexdigest()



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import *
from pixiedust.utils.userPreferences import *
import base64
import requests

access_token_key = 'cf.access_token'
api_base_url = 'api.ng.bluemix.net'
login_base_url = 'login.ng.bluemix.net'
    
@PixieApp
class CFBrowser:


    def get_data_frame(self):
        return self.df
    
    @route(login="true")
    def _login(self):
        # Login
        body = 'grant_type=password&passcode={}'.format(self.passcode)
        url = 'https://{}/UAALoginServerWAR/oauth/token'.format(login_base_url)
        r = requests.post(url, data=body, headers={
            'Authorization': 'Basic Y2Y6',
            'Content-Type': 'application/x-www-form-urlencoded',
            'Accept': 'application/json'
        })
        json = r.json()
        if 'access_token' in json.keys():
            self.access_token = json['access_token']
            setUserPreference(access_token_key, self.access_token)
        self.show_orgs()
        
    def show_orgs(self):
        # Load organizations
        orgs = self.get_orgs(self.access_token)
        options = ""
        for org in orgs:
            options += """<option value="{}">{}</option>""".format(org['metadata']['guid'],org['entity']['name'])
        self._addHTMLTemplateString("""
<div>
  <div class="form-horizontal">
    <div class="form-group">
      <label for="org{{prefix}}" class="control-label col-sm-2">Select an organization:</label>
      <div class="col-sm-5">
        <select class="form-control" id="org{{prefix}}">""" + options + """</select>
      </div>
      <div class="col-sm-1">
        <button type="submit" class="btn btn-primary">Go
          <pd_script>
self.org_id="$val(org{{prefix}})"
self.login="false"
self.select_org="true"
          </pd_script>
        </button>
      </div>
    </div>
  </div>
</div>
""")
        
    @route(select_org="true")
    def _select_org(self):
        spaces = self.get_spaces(self.access_token, self.org_id)
        options = ""
        for space in spaces:
            options += """<option value="{}">{}</option>""".format(space['metadata']['guid'],space['entity']['name'])
        self._addHTMLTemplateString("""
<div>
  <div class="form-horizontal">
    <div class="form-group">
      <label for="org{{prefix}}" class="control-label col-sm-2">Select a space:</label>
      <div class="col-sm-5">
        <select class="form-control" id="space{{prefix}}">""" + options + """</select>
      </div>
      <div class="col-sm-1">
        <button type="submit" class="btn btn-primary">Go
          <pd_script>
self.space_id="$val(space{{prefix}})"
self.select_org="false"
self.select_space="true"
          </pd_script>
        </button>
      </div>
    </div>
  </div>
</div>
""")
        
#     # @route(select_space="true")
#     def _select_space(self):
#         self._addHTMLTemplateString("""
# <div class="row" pixiedust="{{pd_controls|htmlAttribute}}">
#     <div class="form-group col-sm-2" style="padding-right:10px;">
#         <p>
#             <button type="submit" class="btn btn-primary" pd_target="target{{prefix}}">Get Apps
#                 <pd_script>self.print_apps()</pd_script>
#             </button>
#         </p>
#         <p>
#             <button type="submit" class="btn btn-primary" pd_target="target{{prefix}}">Get Services
#                 <pd_script>self.print_services()</pd_script>
#             </button>
#         </p>
#     </div>
#     <div class="form-group col-sm-10">
#         <div id="target{{prefix}}"></div>
#     </div>
# </div>
# """)
#
#     def print_apps(self):
#         apps = self.get_apps(self.access_token, self.space_id)
#         for app in apps:
#             print('<p>{}</p>'.format(app['entity']['name']))

    @route(select_space="true")
    def _select_space(self):
        svcs = self.get_services(self.access_token, self.space_id)
        output = """
<div>
    <div class="form-horizontal">
"""
        for svc in svcs:
            svc_label = self.get_service_label(self.access_token, svc['entity']['service_plan_guid'])
            if svc_label == 'cloudantNoSQLDB':
                svc_keys = self.get_service_keys(self.access_token, svc['metadata']['guid'])
                for svc_key in svc_keys:
                    svc_key_entity = svc_key['entity']
                    if 'credentials' in svc_key_entity.keys():
                        credentials_str = json.dumps(svc_key_entity['credentials'])
                        credentials_str = credentials_str.replace('"', '\\"')
                        output += """
<div class="form-group">
    <div class="col-sm-2"></div>
    <div class="col-sm-5">
        <b>""" + svc['entity']['name'] + """</b><br>
        """ + svc_key_entity['credentials']['host'] + """<br>
        <button type="submit" class="btn btn-primary" data-dismiss="modal">Select
            <pd_script>self.service_name=\"""" + svc['entity']['name'].replace('"', '\\"') + """\"
self.credentials=\"""" + credentials_str + """\"
self.select_space="false"
self.select_credentials="true"</pd_script>
        </button>
    </div>
</div>"""
        return output

    @route(select_credentials="true")
    def _select_credentials(self):
        return self.selectBluemixCredentials(self.service_name, self.credentials)

    def is_valid_access_token(self, access_token):
        url = 'https://{}/v2/organizations'.format(api_base_url)
        authHeader = 'Bearer {}'.format(access_token)
        r = requests.get(url, headers={
            'Authorization': authHeader,
            'Accept': 'application/json'
        })
        return r.status_code == 200
        
    def get_orgs(self, access_token):
        url = 'https://{}/v2/organizations'.format(api_base_url)
        authHeader = 'Bearer {}'.format(access_token)
        r = requests.get(url, headers={
            'Authorization': authHeader,
            'Accept': 'application/json'
        })
        json = r.json()
        return json['resources']
        
    def get_spaces(self, access_token, org_id):
        url = 'https://{}/v2/organizations/{}/spaces'.format(api_base_url, org_id)
        authHeader = 'Bearer {}'.format(access_token)
        r = requests.get(url, headers={
            'Authorization': authHeader,
            'Accept': 'application/json'
        })
        json = r.json()
        return json['resources']
        
    def get_apps(self, access_token, space_id):
        url = 'https://{}/v2/apps?q=space_guid:{}'.format(api_base_url, space_id)
        authHeader = 'Bearer {}'.format(access_token)
        r = requests.get(url, headers={
            'Authorization': authHeader,
            'Accept': 'application/json'
        })
        json = r.json()
        return json['resources']
        
    def get_services(self, access_token, space_id):
        url = 'https://{}/v2/service_instances?q=space_guid:{}'.format(api_base_url, space_id)
        authHeader = 'Bearer {}'.format(access_token)
        r = requests.get(url, headers={
            'Authorization': authHeader,
            'Accept': 'application/json'
        })
        json = r.json()
        return json['resources']
    
    def get_service_keys(self, access_token, service_id):
        url = 'https://{}/v2/service_keys?q=service_instance_guid:{}'.format(api_base_url, service_id)
        authHeader = 'Bearer {}'.format(access_token)
        r = requests.get(url, headers={
            'Authorization': authHeader,
            'Accept': 'application/json'
        })
        json = r.json()
        return json['resources']
    
    def get_service_label(self, access_token, service_plan_id):
        # Load the service plan
        url = 'https://{}/v2/service_plans/{}'.format(api_base_url, service_plan_id)
        authHeader = 'Bearer {}'.format(access_token)
        r = requests.get(url, headers={
            'Authorization': authHeader,
            'Accept': 'application/json'
        })
        json = r.json()
        if 'entity' in json.keys():
            service_id = json['entity']['service_guid']
        else:
            return "NO PLAN FOUND"
        # Load the service
        url = 'https://{}/v2/services/{}'.format(api_base_url, service_id)
        path = '/v2/services/{}'.format(service_id)
        r = requests.get(url, headers={
            'Authorization': authHeader,
            'Accept': 'application/json'
        })
        json = r.json()
        return json['entity']['label']
        
    def cloudant_all_dbs(self, host, username, password):
        url = 'https://{}/_all_dbs'.format(host)
        r = requests.get(url, headers={
            'Authorization': 'Basic {}'.format(base64.b64encode('{}:{}'.format(username, password))),
            'Accept': 'application/json'
        })
        return r.json()

    @route()
    def startBrowsingBM(self):
        access_token_valid = False
        access_token = getUserPreference(access_token_key)
        if access_token is not None:
            access_token_valid = self.is_valid_access_token(access_token)
        if access_token_valid:
            self.access_token = access_token
            self.show_orgs()
        else:
            return """
<div>
  <div class="form-horizontal">
    <div class="form-group">
        <div class="col-sm-2"></div>
        <div class="col-sm-5">
            <a href="https://login.ng.bluemix.net/UAALoginServerWAR/passcode" target="_blank">
            Click here to get your one-time passcode
            </a>
        </div>
    </div>
    <div class="form-group">
      <label for="passcode{{prefix}}" class="control-label col-sm-2">Passcode:</label>
      <div class="col-sm-5">
        <input type="text" class="form-control" id="passcode{{prefix}}">
      </div>
      <div class="col-sm-1">
        <button type="submit" class="btn btn-primary">
          Go
          <pd_script>
self.passcode="$val(passcode{{prefix}})"
self.login="true"
          </pd_script>
        </button>
      </div>
    </div>
  </div>
</div>
"""


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Inherited from maven-artifact https://github.com/hamnis/maven-artifact
# -------------------------------------------------------------------------------
import json
import os
import re
import requests
import shutil
import string
import subprocess
import sys
import tarfile
import tempfile
from ipykernel.kernelspec import KernelSpecManager, write_kernel_spec
from jupyter_client.kernelspec import NoSuchKernel
from jupyter_client.manager import KernelManager
from jupyter_client.kernelspecapp  import InstallKernelSpec
from six.moves import input
from traitlets.config.application import Application
from lxml import etree
import nbformat

class PixiedustInstall(InstallKernelSpec):
    def __init__(self, **kwargs):
        super(PixiedustInstall, self).__init__(**kwargs)
        self.pixiedust_home = None
        self.spark_home = None
        self.spark_download_versions = ['1.6.3', '2.0.2', '2.1.0']
        self.spark_download_urls = {
            '1.6.3': 'http://d3kbcqa49mib13.cloudfront.net/spark-1.6.3-bin-hadoop2.6.tgz',
            '2.0.2': 'http://d3kbcqa49mib13.cloudfront.net/spark-2.0.2-bin-hadoop2.7.tgz',
            '2.1.0': 'http://d3kbcqa49mib13.cloudfront.net/spark-2.1.0-bin-hadoop2.7.tgz'
        }
        self.scala_home = None
        self.scala_download_urls = {
            '2.10': 'http://downloads.lightbend.com/scala/2.10.6/scala-2.10.6.tgz',
            '2.11': 'http://downloads.lightbend.com/scala/2.11.8/scala-2.11.8.tgz'
        }

    def parse_command_line(self, argv):
        silent = "--silent" in argv
        silent_spark_version = None
        if silent:
            argv.remove("--silent")
            arg_str = None
            for i, arg in enumerate(argv):
                if arg.startswith("--spark"):
                    arg_str = arg
                    silent_spark_version = arg_str[len("--spark")+1:]
                    if silent_spark_version not in self.spark_download_versions:
                        print("Invalid Spark version {}".format(silent_spark_version))
                        self.exit(1)
                    break
            if arg_str:
                argv.remove(arg_str)
        super(InstallKernelSpec, self).parse_command_line(argv)

        self.pixiedust_home = os.environ.get("PIXIEDUST_HOME", "{}{}pixiedust".format(os.path.expanduser('~'), os.sep))
        if silent:
            answer = 'y'
        else:
            answer = self.confirm(
                "Step 1: PIXIEDUST_HOME: {0}".format(self.pixiedust_home)
            )

        if answer != 'y':
            self.pixiedust_home = input(self.hilite("Please enter a PIXIEDUST_HOME location: "))
            if not os.path.exists(self.pixiedust_home):
                create = self.confirm("Directory {0} does not exist".format(self.pixiedust_home), "Create")
                if create != 'y':
                    self.exit(1)
                else:
                    os.makedirs(self.pixiedust_home)

        self.pixiedust_bin = os.path.join(self.pixiedust_home, "bin")
        download_spark = False
        if silent:
            self.spark_home = os.environ.get("SPARK_HOME", None)
            if not self.spark_home:
                download_spark = True
                self.spark_home = "{}{}spark".format(os.path.expanduser('~'), os.sep)
        else:
            first_prompt = True
            while True:
                self.spark_home = os.environ.get("SPARK_HOME", self.getFirstDir(os.path.join(self.pixiedust_bin, "spark" )))
                if self.spark_home:
                    answer = self.confirm(
                        "Step 2: SPARK_HOME: {0}".format(self.spark_home)
                    )
                    if answer != 'y':
                        self.spark_home = input(self.hilite("Step 2: Please enter a SPARK_HOME location: "))
                else:
                    if first_prompt:
                        first_prompt = False
                        prompt = "Step 2: Please enter a SPARK_HOME location: "
                    else:
                        prompt = "Please enter a SPARK_HOME location: "
                    self.spark_home = input(self.hilite(prompt))
                while self.spark_home.rfind(os.sep) == len(self.spark_home) - 1:
                    self.spark_home = self.spark_home[0:len(self.spark_home)-1]
                if not os.path.exists(self.spark_home):
                    create = self.confirm("Directory {0} does not exist".format(self.spark_home), "Create")
                    if create != 'y':
                        continue
                    else:
                        os.makedirs(self.spark_home)
                        download_spark = True
                        break
                elif not os.path.exists('{}{}bin{}pyspark'.format(self.spark_home, os.sep, os.sep)):
                    download = self.confirm("Directory {0} does not contain a valid SPARK install".format(self.spark_home), "Download Spark")
                    if download == 'y':
                        download_spark = True
                        break
                else:
                    break

        if download_spark:
            self.download_spark(silent, silent_spark_version)

        scala_version = None
        spark_version = self.get_spark_version()
        if spark_version is None:
            print("Unable to obtain Spark version")
            self.exit(1)
        elif spark_version[0] == 1:
            scala_version = '2.10'  # Spark 1.x = Scala 2.10
        elif spark_version[0] == 2:
            scala_version = '2.11'  # Spark 2.x = Scala 2.11
        else:
            print("Invalid Spark version {}".format(spark_version))
            self.exit(1)

        #download spark-cloudant
        sparkCloudantUrl = "https://github.com/cloudant-labs/spark-cloudant/releases/download/v1.6.4/cloudant-spark-v1.6.4-167.jar"
        if spark_version[0] == 2:
            sparkCloudantUrl = "https://github.com/cloudant-labs/spark-cloudant/releases/download/v2.0.0/cloudant-spark-v2.0.0-185.jar"

        try:
            self.sparkCloudantPath = self.downloadFileToDir( sparkCloudantUrl, targetDir = self.pixiedust_bin )
            print("downloaded spark cloudant jar: {0}".format(self.sparkCloudantPath))
        except Exception as e:
            print("Error downloading Cloudant Jar: {}. Install will continue without the spark Cloudant connector".format(e))

        #download spark csv connector if in spark 1.6
        self.sparkCSVPath = None
        self.commonsCSVPath = None
        if spark_version[0] == 1:
            try:
                self.sparkCSVPath = self.downloadPackage("com.databricks:spark-csv_2.10:1.5.0")
                self.commonsCSVPath = self.downloadPackage("org.apache.commons:commons-csv:0")
            except Exception as e:
                print("Error downloading csv connector Jar: {}. Install will continue without it".format(e))

        download_scala = False
        if silent:
            self.scala_home = os.environ.get("SCALA_HOME", None)
            if not self.scala_home:
                download_scala = True
                self.scala_home = "{}{}scala".format(os.path.expanduser('~'), os.sep)
        else:
            first_prompt = True
            while True:
                self.scala_home = os.environ.get("SCALA_HOME", self.getFirstDir(os.path.join(self.pixiedust_bin, "scala" )))
                if self.scala_home:
                    answer = self.confirm(
                        "Step 3: SCALA_HOME: {0}".format(self.scala_home)
                    )
                    if answer != 'y':
                        self.scala_home = input(self.hilite("Step 3: Please enter a SCALA_HOME location: "))
                else:
                    if first_prompt:
                        first_prompt = False
                        prompt = "Step 3: Please enter a SCALA_HOME location: "
                    else:
                        prompt = "Please enter a SCALA_HOME location: "
                    self.scala_home = input(self.hilite(prompt))
                while self.scala_home.rfind(os.sep) == len(self.scala_home) - 1:
                    self.scala_home = self.scala_home[0:len(self.scala_home)-1]
                if not os.path.exists(self.scala_home):
                    create = self.confirm("Directory {0} does not exist".format(self.scala_home), "Create")
                    if create != 'y':
                        continue
                    else:
                        os.makedirs(self.scala_home)
                        download_scala = True
                        break
                elif not os.path.exists('{}{}bin{}scala'.format(self.scala_home, os.sep, os.sep)):
                    download = self.confirm(
                        "Directory {0} does not contain a valid SCALA install".format(self.scala_home),
                        "Download Scala"
                    )
                    if download == 'y':
                        download_scala = True
                        break
                else:
                    installed_scala_version = self.get_scala_version()
                    if not installed_scala_version or (str(installed_scala_version[0]) + '.' + str(installed_scala_version[1])) != scala_version:
                        print("A different version of Scala {0} is already installed in this directory.".format(installed_scala_version))
                        continue
                    else:
                        break

        if download_scala:
            self.download_scala(scala_version)

        self.kernelName = "Python with Pixiedust (Spark {}.{})".format(spark_version[0], spark_version[1])
        if silent:
            answer = 'y'
        else:
            answer = self.confirm(
                "Step 4: Kernel Name: {0}".format(self.kernelName)
            )
        if answer != 'y':
            self.kernelName = input(self.hilite("Please enter a Kernel Name: "))

        try:
            km = KernelManager(kernel_name=self.kernelName)
            km.kernel_spec
            answer = self.confirm(
                "Kernel '{0}' already exists".format(self.kernelName), "Override"
            )
            if answer != 'y':
                self.exit(1)
        except NoSuchKernel:
            pass

        dest = self.createKernelSpec()

        self.pixiedust_notebooks_dir = os.path.join( self.pixiedust_home, "notebooks")
        if not os.path.isdir(self.pixiedust_notebooks_dir):
            os.makedirs(self.pixiedust_notebooks_dir)
        print("Downloading intro notebooks into {}".format(self.pixiedust_notebooks_dir))
        self.downloadIntroNotebooks()

        print("\n\n{}".format("#"*100))
        print("#\tCongratulations: Kernel {0} was successfully created in {1}".format(self.kernelName, dest))
        print("#\tYou can start the Notebook server with the following command:")
        print("#\t\t{}".format(
            self.hilite("jupyter notebook {}".format(self.pixiedust_notebooks_dir))
        ))
        print("{}".format("#"*100))

    def getFirstDir(self, parentLoc):
        if not os.path.isdir(parentLoc):
            return parentLoc
        dirs = os.listdir(parentLoc)
        return parentLoc if len(dirs) == 0 else os.path.join(parentLoc, dirs[0])

    def downloadFileToDir( self, url, targetDir ):
        index = url.rfind('/')
        fileName = url[index+1:] if index >= 0 else url
        localPath = os.path.join(targetDir, fileName)
        if not os.path.isfile(localPath ):
            try:
                with open(localPath, "wb") as targetFile:
                    self.download_file(url, targetFile = targetFile)
            except:
                os.remove(localPath)
                raise
        return localPath

    def downloadPackage(self, packageId):
        if packageId.startswith("http://") or packageId.startswith("https://") or packageId.startswith("file://"):
            #direct download
            return self.downloadFileToDir( packageId, targetDir=self.pixiedust_bin)
        else:
            #assume maven id, resolve it
            parts = packageId.split(":")
            bases = ["http://repo1.maven.org/maven2", "http://dl.bintray.com/spark-packages/maven"]
            response = None
            for base in bases:
                url = None
                if parts[2] == '0':
                    #resolve version
                    metadataUrl = "{0}/{1}/maven-metadata.xml".format(base, "/".join([parts[0].replace(".", "/"), parts[1]]))
                    r = requests.get(metadataUrl, stream=True)
                    if r.ok:
                        parts[2] = etree.parse(r.raw).xpath("/metadata/versioning/versions/version[last()]/text()")[0]
                if parts[2] != '0':
                    url = "{0}/{1}/{2}-{3}.jar".format(
                        base, 
                        "/".join([parts[0].replace(".", "/"), parts[1], parts[2]]),
                        parts[1],
                        parts[2]
                    )
                    response = requests.get(url, stream=True)
                    if response.ok:
                        break

            if response is None or not response.ok:
                raise Exception("Unable to resolve package {}".format(packageId))

            return self.downloadFileToDir( url, targetDir=self.pixiedust_bin)

    def downloadIntroNotebooks(self):
        #download the Intro.txt file that contains all the notebooks to download
        response = requests.get("https://github.com/ibm-cds-labs/pixiedust/raw/master/notebook/Intro.txt")
        if not response.ok:
            raise Exception("Unable to read the list of Intro Notebooks")

        notebookNames = response.content.decode().split("\n")
        introNotebooksUrls = [
            "https://github.com/ibm-cds-labs/pixiedust/raw/master/notebook/" + n for n in notebookNames if n != ""
        ]
        for url in introNotebooksUrls:
            print("...{0}".format(url))
            try:
                path = self.downloadFileToDir(url, targetDir=self.pixiedust_notebooks_dir)
                #update kernel name and display_name
                f = open(path, 'r')
                contents = f.read()
                f.close()
                nb=nbformat.reads(contents, as_version=4)
                nb.metadata.kernelspec.name=self.kernelInternalName
                nb.metadata.kernelspec.display_name = self.kernelName
                f = open(path, 'w')
                f.write(json.dumps(nb))
                f.close()
                print("\033[F\033[F")
                print("...{0} : {1}".format(url, self.hilite("done")))
            except Exception as e:
                print("\033[F\033[F")
                print("...{0} : {1}".format(url, self.hilite("Error {}".format(e))))
                

    def get_spark_version(self):
        pyspark = "{}{}bin{}pyspark".format(self.spark_home, os.sep, os.sep)
        pyspark_out = subprocess.check_output([pyspark, "--version"], stderr=subprocess.STDOUT).decode("utf-8")
        match = re.search(".*version[^0-9]*([0-9]*[^.])\.([0-9]*[^.])\.([0-9]*[^.]).*", str(pyspark_out))
        if match and len(match.groups()) > 2:
            return int(match.group(1)), int(match.group(2))
        else:
            return None

    def download_spark(self, silent, silent_spark_version):
        while True:
            spark_default_version = self.spark_download_versions[len(self.spark_download_versions)-1]
            spark_download_versions_str = ', '.join(self.spark_download_versions) + ' [{}]'.format(spark_default_version)
            if silent:
                if silent_spark_version:
                    spark_version = silent_spark_version
                else:
                    spark_version = spark_default_version
            else:
                spark_version = input(
                    self.hilite("What version would you like to download? {}: ".format(spark_download_versions_str))
                )
            if len(spark_version.strip()) == 0:
                spark_version = self.spark_download_versions[len(self.spark_download_versions)-1]
            elif spark_version not in self.spark_download_versions:
                print("{0} is not a valid version".format(self.spark_home))
                continue
            spark_download_url = self.spark_download_urls[spark_version]
            spark_download_file = spark_download_url[spark_download_url.rfind('/')+1:spark_download_url.rfind('.')]
            print("SPARK_HOME will be set to {}{}{}".format(self.spark_home, os.sep, spark_download_file))
            print("Downloading Spark {}".format(spark_version))
            temp_file = self.download_file(spark_download_url)
            print("Extracting Spark {} to {}".format(spark_version, self.spark_home))
            self.extract_temp_file(temp_file, self.spark_home)
            self.delete_temp_file(temp_file)
            self.spark_home = "{}{}{}".format(self.spark_home, os.sep, spark_download_file)
            os.environ["SPARK_HOME"] = self.spark_home
            break

    def get_scala_version(self):
        scala = "{}{}bin{}scala".format(self.scala_home, os.sep, os.sep)
        try:
            scala_out = subprocess.check_output([scala, "-version"], stderr=subprocess.STDOUT).decode("utf-8")
        except subprocess.CalledProcessError as cpe:
            scala_out = cpe.output
        match = re.search(".*version[^0-9]*([0-9]*[^.])\.([0-9]*[^.])\.([0-9]*[^.]).*", str(scala_out))
        if match and len(match.groups()) > 2:
            return int(match.group(1)), int(match.group(2))
        else:
            return None

    def download_scala(self, scala_version):
        while True:
            scala_download_url = self.scala_download_urls[scala_version]
            scala_download_file = scala_download_url[scala_download_url.rfind('/')+1:scala_download_url.rfind('.')]
            print("SCALA_HOME will be set to {}/{}".format(self.scala_home, scala_download_file))
            print("Downloading Scala {}".format(scala_version))
            temp_file = self.download_file(scala_download_url)
            print("Extracting Scala {} to {}".format(scala_version, self.scala_home))
            self.extract_temp_file(temp_file, self.scala_home)
            self.delete_temp_file(temp_file)
            self.scala_home = "{}{}{}".format(self.scala_home, os.sep, scala_download_file)
            os.environ["SCALA_HOME"] = self.scala_home
            break

    @staticmethod
    def download_file(url, suffix=None, targetFile = None):
        if suffix is None:
            suffix = url[url.rfind('.'):]

        if targetFile is None:
            targetFile = tempfile.NamedTemporaryFile(suffix=suffix)

        response = requests.get(url, stream=True)
        if not response.ok:
            raise Exception("{}".format(response.status_code))

        total_bytes = int(response.headers.get("content-length", 0))
        bytes_read = 0
        for chunk in response.iter_content(chunk_size=1048576):
            targetFile.write(chunk)
            targetFile.flush()
            bytes_read += len(chunk)
            if total_bytes <= bytes_read:
                total_bytes = bytes_read
            print(" {} %".format(int(((bytes_read*1.0)/total_bytes)*100)))
            sys.stdout.write("\033[F")
            sys.stdout.flush()
        print("      ")
        sys.stdout.write("\033[F")
        sys.stdout.flush()
        return targetFile

    @staticmethod
    def delete_temp_file(temp_file):
        temp_file.close()

    @staticmethod
    def extract_temp_file(temp_file, path):
        tar = tarfile.open(temp_file.name, "r:gz")
        for i, member in enumerate(tar.getmembers()):
            tar.extract(member, path)
            print(" {} %".format(int(((i*1.0)/len(tar.getmembers()))*100)))
            sys.stdout.write("\033[F")
            sys.stdout.flush()
        print("      ")
        sys.stdout.write("\033[F")
        sys.stdout.flush()
        tar.close()

    def createKernelSpec(self):
        try:
            libFiles = os.listdir("{0}/python/lib".format(self.spark_home))
            py4j_zip = list(filter( lambda filename: "py4j" in filename, libFiles))[0]
        except:
            print('Unable to find py4j in SPARK_HOME: {0}'.format(self.spark_home))
            self.exit(1)

        overrides={
            "display_name": self.kernelName,
            "argv": [
                "python",
                "-m",
                "ipykernel",
                "-f",
                "{connection_file}"
            ],
            "env": {
                "PIXIEDUST_HOME": self.pixiedust_home,
                "SCALA_HOME": "{0}".format(self.scala_home),
                "SPARK_HOME": "{0}".format(self.spark_home),
                "PYTHONPATH": "{0}/python/:{0}/python/lib/{1}".format(self.spark_home, py4j_zip),
                "PYTHONSTARTUP": "{0}/python/pyspark/shell.py".format(self.spark_home),
                "PYSPARK_SUBMIT_ARGS": "--jars {0} --driver-class-path {1} --master local[10] pyspark-shell".format(
                    self.sparkCloudantPath,
                    ":".join([x for x in [self.pixiedust_home + "/data/libs/*" ,self.sparkCSVPath,self.commonsCSVPath] if x is not None]),
                ),
                "SPARK_DRIVER_MEMORY":"10G",
                "SPARK_LOCAL_IP":"127.0.0.1"
            }
        }
        path = write_kernel_spec(overrides=overrides)
        self.kernelInternalName = ''.join(ch for ch in self.kernelName if ch.isalnum()).lower()
        print("self.kernelInternalName {}".format(self.kernelInternalName))
        dest = KernelSpecManager().install_kernel_spec(path, kernel_name=self.kernelInternalName, user=True)
        # cleanup afterward
        shutil.rmtree(path)
        return dest

    def confirm(self, message, action="Keep"):
        answer = input(self.hilite(message) + "\n\t" + action + " y/n [y]? ")
        return 'y' if answer == '' else answer

    def hilite(self, message):
        return '\x1b[%sm%s\x1b[0m' % (';'.join(['32', '1']), message)

    def start(self):
        pass


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.display import CellHandshake
from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.utils import Logger
from ..baseChartDisplay import BaseChartDisplay
from six import with_metaclass
from abc import abstractmethod, ABCMeta
from bokeh.plotting import figure, output_notebook
from bokeh.models.tools import *
from bokeh.io import notebook_div
import pkg_resources

@PixiedustRenderer(rendererId="bokeh")
@Logger()
class BokehBaseDisplay(with_metaclass(ABCMeta, BaseChartDisplay)):
    CellHandshake.addCallbackSniffer( lambda: "{'nostore_bokeh':!!window.Bokeh}")

    #get the bokeh version
    try:
        bokeh_version = pkg_resources.get_distribution("bokeh").parsed_version._version.release
    except:
        bokeh_version = None
        self.exception("Unable to get bokeh version")

    def __init__(self, options, entity, dataHandler=None):
        super(BokehBaseDisplay,self).__init__(options,entity,dataHandler)
        #no support for orientation
        self.no_orientation = True
    
    """
    Default implementation for creating a chart object.
    """
    def createBokehChart(self):
        return figure()

    def getPreferredOutputWidth(self):
        return super(BokehBaseDisplay,self).getPreferredOutputWidth() * 0.92

    def doRenderChart(self):
        def genMarkup(chartFigure):
            return self.env.from_string("""
                    {0}
                    {{%for message in messages%}}
                        <div>{{{{message}}}}</div>
                    {{%endfor%}}
                """.format(chartFigure)
            ).render(messages=self.messages)

        if BokehBaseDisplay.bokeh_version < (0,12):
            raise Exception("""
                <div>Incorrect version of Bokeh detected. Expected {0}, got {1}</div>
                <div>Please upgrade by using the following command: <b>!pip install --user --upgrade bokeh</b></div>
            """.format((0,12), BokehBaseDisplay.bokeh_version))
        clientHasBokeh = self.options.get("nostore_bokeh", "false") == "true"
        if not clientHasBokeh:          
            output_notebook(hide_banner=True)
        charts = self.createBokehChart()

        if not isinstance(charts, list):
            charts.add_tools(ResizeTool())
            #bokeh 0.12.5 has a non backward compatible change on the title field. It is now of type Title
            #following line is making sure that we are still working with 0.12.4 and below
            if hasattr(charts, "title") and hasattr(charts.title, "text"):
                charts.title.text = self.options.get("title", "")
            else:
                charts.title = self.options.get("title", "")
            charts.plot_width = int(self.getPreferredOutputWidth() - 10 )
            charts.plot_height = int(self.getPreferredOutputHeight() - 10  )
            charts.grid.grid_line_alpha=0.3
            return genMarkup(notebook_div(charts))
        else:
            from bokeh.layouts import gridplot
            ncols = 2
            nrows = len(charts)/2 + len(charts)%2
            
            w = self.getPreferredOutputWidth()/ncols if len(charts) > 1 else self.getPreferredOutputWidth()
            h = w * self.getHeightWidthRatio() if len(charts) > 1 else self.getPreferredOutputHeight()
            for chart in charts:
                chart.plot_width = int(w - 5)
                chart.plot_height = int (h - 5)

            return genMarkup(notebook_div(gridplot(charts, ncols=ncols, nrows=nrows)))


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.colorManager import Colors
from .bokehBaseDisplay import BokehBaseDisplay
from pixiedust.utils import Logger
import pandas as pd
import numpy
import bokeh.plotting as gridplot
import sys

try:
    from bkcharts import Bar
    from bkcharts.operations import blend
    from bkcharts.attributes import  CatAttr
except ImportError:
    from bokeh.charts import Bar
    from bokeh.charts.operations import blend
    from bokeh.charts.attributes import  CatAttr

@PixiedustRenderer(id="barChart")
@Logger()
class BarChartRenderer(BokehBaseDisplay):
    def getSubplotHSpace(self):
        return 0.5

    def isSubplot(self):
        return self.options.get("charttype", "grouped") == "subplots"

    def getExtraFields(self):
        if not self.isSubplot() and len(self.getValueFields())>1:
            #no clusterby if we are grouped and multiValueFields
            return []
    
        clusterby = self.options.get("clusterby")
        return [clusterby] if clusterby is not None else []

    def createBokehChart(self):
        keyFields = self.getKeyFields()
        valueFields = self.getValueFields()
        clusterby = self.options.get("clusterby")
        stacked = self.options.get("charttype", "grouped") == "stacked"
        subplots = self.isSubplot()
        workingPDF = self.getWorkingPandasDataFrame().copy()

        def convertPDFDate(df, col):
            #Bokeh doesn't support datetime as index in Bar chart. Convert to String
            if len(keyFields) == 1:
                dtype = df[col].dtype.type if col in df else None
                if numpy.issubdtype(dtype, numpy.datetime64):
                    dateFormat = self.options.get("dateFormat", None)
                    try:
                        df[col] = df[col].apply(lambda x: str(x).replace(':','-') if dateFormat is None else x.strftime(dateFormat))
                    except:
                        self.exception("Error converting dateFormat {}".format(dateFormat))
                        df[col] = df[col].apply(lambda x: str(x).replace(':','-'))

        for index, row in workingPDF.iterrows():
            for k in keyFields:
                if isinstance(row[k], str if sys.version >= '3' else basestring):
                    row[k] = row[k].replace(':', '.')
            workingPDF.loc[index] = row

        charts=[]
        def goChart(label, stack_or_group, values, ylabel=None, color=None):
            convertPDFDate(workingPDF, keyFields[0])
            if ylabel is None:
                ylabel=values
            label=label if isinstance(label, (list, tuple)) else [label]
            if stacked:
                charts.append( Bar(workingPDF, label=CatAttr(columns=label, sort=False), stack=stack_or_group, color=color, values=values, legend=self.showLegend(), ylabel=ylabel))
            else:
                charts.append( Bar(workingPDF, label=CatAttr(columns=label, sort=False), group=stack_or_group, color=color, values=values, legend=self.showLegend(), ylabel=ylabel))

        if clusterby is not None and (subplots or len(valueFields)<=1):
            subplots = subplots if len(valueFields)==1 or subplots else False
            if subplots:
                for j, valueField in enumerate(valueFields):
                    pivot = workingPDF.pivot(
                        index=keyFields[0], columns=clusterby, values=valueField
                    )
                    for i,col in enumerate(pivot.columns[:10]): #max 10
                        data = pd.DataFrame({'values':pivot[col].values, 'names': pivot.index.values})
                        convertPDFDate(data, 'names')
                        if subplots:                        
                            charts.append( 
                                Bar(data, label=CatAttr(columns=['names'], sort=False), color = Colors.hexRGB( 1.*i/2 ), values='values', ylabel=valueField, legend=False, 
                                    title="{0} = {1}".format(clusterby, pivot.columns[i])
                                )
                            )
            else:
                goChart( keyFields[0], clusterby, valueFields[0])
        else:
            if subplots:
                for i,valueField in enumerate(valueFields):
                    goChart( keyFields[0], None, valueField, color=Colors.hexRGB( 1.*i/2 ))
            else:
                if len(valueFields) > 1:
                    series = '_'.join(valueFields)
                    values = blend(*valueFields, name=series.replace('_', ','), labels_name=series)
                else:
                    series = False
                    values = valueFields[0]
                goChart(keyFields, series, values, ylabel=','.join(valueFields))

            if clusterby is not None:
                self.addMessage("Warning: 'Cluster By' ignored when grouped option with multiple Value Fields is selected")
        return charts


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from abc import ABCMeta,abstractmethod
from IPython.display import display as ipythonDisplay, HTML, Javascript
from pixiedust.utils import Logger
from pixiedust.utils.template import *
import sys
import uuid
from collections import OrderedDict
import time
import re
import pixiedust
from six import iteritems, with_metaclass
from functools import reduce
import json

myLogger = pixiedust.getLogger(__name__)

handlers=[]
systemHandlers=[]
defaultHandler=None
globalMenuInfos={}

"""
Registry of Action categories
"""
class ActionCategories(object):
    CAT_INFOS = OrderedDict([
        ("Table", {"title": "Table", "icon-class": "fa-table"}),
        ("Chart", {"title": "Chart", "icon-class": "fa-line-chart"}),
        ("Map",   {"title": "Map", "icon-class": "fa-map"}),
        ("Graph", {"title": "Graph", "icon-class": "fa-share-alt"}),
        ("Download", {"title":  "Stash dataset", "icon-class": "fa-cloud-download", "pos": 100})
    ])
    
    @staticmethod
    def sort():
        ActionCategories.CAT_INFOS = OrderedDict(sorted(iteritems(ActionCategories.CAT_INFOS), key=lambda item: item[1]["pos"] if "pos" in item[1] else 0))

def registerDisplayHandler(handlerMetadata, isDefault=False, system=False):
    global defaultHandler
    if isDefault and defaultHandler is None:
        defaultHandler=handlerMetadata
    if system:
        systemHandlers.append(handlerMetadata)
    else:
        handlers.append(handlerMetadata)

    #Add the categories
    cat = None
    for cat in handlerMetadata.createCategories():
        myLogger.debug("Adding category {0} to registry".format(str(cat)))
        if "id" not in cat:
            myLogger.error("Invalid category {0}".format(str(cat)))
        else:
            #fix the icon-path if available
            if "icon-path" in cat and ":" not in cat["icon-path"]:
                cat["icon-path"] = handlerMetadata.__module__ + ":" + cat["icon-path"]
            ActionCategories.CAT_INFOS[cat["id"]] = cat
    if cat is not None:
        #resort the registry by position
        ActionCategories.sort()
    
def getSelectedHandler(options, entity, dataHandler):
    if "cell_id" not in options:
        #No cellid, trigger handshake with the browser to get the cellId
        return CellHandshakeMeta()
    elif options.get('runInDialog', 'false') == 'true':
        #we are running in a dialog
        return RunInDialogMeta()
        
    handlerId=options.get("handlerId")
    if handlerId is not None:
        if handlerId in globalMenuInfos:
            return globalMenuInfos[handlerId]['handler']
        else:
            #we need to find it
            def findHandlerId(e):
                for handler in (handlers+systemHandlers):
                    for menuInfo in handler.getMenuInfo(e, dataHandler):
                        if handlerId in globalMenuInfos:
                            return globalMenuInfos[handlerId]['handler']

            retValue = findHandlerId(entity)            
            if retValue:
                return retValue

    #If we're here, then either no handlerId was specified or if it was, it was not found
    if handlerId is not None:
        myLogger.debug("Unable to resolve handlerId {} for entity {}. Trying to find a suitable one".format(handlerId, entity) )

    if defaultHandler is not None and len(defaultHandler.getMenuInfo(entity, dataHandler))>0:
        return defaultHandler
    #get the first handler that can render this object
    for handler in handlers:
        menuInfos = handler.getMenuInfo(entity, dataHandler)
        if ( menuInfos is not None and len(menuInfos)>0 ):
            return handler
    #we didn't find any, return the first
    myLogger.debug("Didn't find any handler for {0}".format(handlerId))
    return UnknownEntityMeta()

"""
misc helper functions
"""
def safeCompare(entity1, entity2):
    try:
        return entity1 == entity2
    except:
        return False

"""
PixieDust display class decorator
"""
class PixiedustDisplayMeta(object):
    def __init__(self, **kwArgs):
        self.keywordArgs = kwArgs

    def __call__(self, cls, *args, **kwargs):
        registerDisplayHandler(cls(), **self.keywordArgs)
        return cls
"""
For backward compatibility only. Deprecated
"""
class PixiedustDisplay(PixiedustDisplayMeta):
    pass

def addId(func):
    def wrapper(*args,**kwargs):
        global globalMenuInfos
        menuInfos = func(*args, **kwargs)
        for menuInfo in menuInfos:
            if 'id' not in menuInfo:
                raise Exception("MenuInfo json must have a unique id")            
            globalMenuInfos[menuInfo['id']]=menuInfo
            menuInfo['handler']=args[0]
        return menuInfos
    return wrapper

class DisplayHandlerMeta(with_metaclass(ABCMeta)):
    @abstractmethod
    @addId
    def getMenuInfo(self, entity, dataHandler):
        pass
    @abstractmethod
    def newDisplayHandler(self,options,entity):
        pass

    def createCategories(self):
        return []

@Logger()
class Display(with_metaclass(ABCMeta)):

    #global jinja2 Environment
    env = PixiedustTemplateEnvironment()
    
    def __init__(self, options, entity, dataHandler=None):
        self.entity=entity
        self.options=options
        self.dataHandler=dataHandler
        self.html=""
        self.scripts=list()
        self.noChrome="handlerId" in options and "showchrome" not in options
        self.addProfilingTime = False
        self.executionTime=None
        self.extraTemplateArgs={}

    def getBooleanOption(self, key, defValue):
        value = self.options.get(key, None)
        if value is None:
            return defValue
        return value == "true"

    def getDPI(self):
        return int(self.options.get("nostore_dpi", 96))

    """
        return chart height to width ration
    """
    def getHeightWidthRatio(self):
        return 0.75

    """
        return chart width scale factor
    """
    def getWidthScaleFactor(self):
        return 0.8 if "no_margin" not in self.options else 1.0

    def getPreferredOutputWidth(self):
        sizeratio = float(self.options.get('chartsize', 100)) / 100
        return float(self.options.get("nostore_cw", 1000)) * sizeratio * self.getWidthScaleFactor()

    def getPreferredOutputHeight(self):
        ch = self.options.get("nostore_ch", None)
        if ch is not None:
            return float(ch)
        else:
            return float(self.getPreferredOutputWidth() * self.getHeightWidthRatio())


    def _getTemplateArgs(self, **kwargs):
        args = {
            "this":self, 
            "entity":self.entity, 
            "prefix":self.getPrefix(),
            "module":self.__module__,
            "pd_controls": json.dumps({
                "prefix": self.getPrefix(),
                "command": self._genDisplayScript(menuInfo=kwargs.get("menuInfo", None) ),
                "options": self.options
            })
        }

        args.update(self.extraTemplateArgs)
        if kwargs:
            args.update(kwargs)
        return args

    def renderTemplate(self, templateName, **kwargs):
        return self.env.getTemplate(templateName).render(self._getTemplateArgs(**kwargs))

    """
        check if pixiedust object is already installed on the client
    """
    def _checkPixieDustJS(self):
        if self.options.get("nostore_pixiedust", "false") != "true":
            self.options["nostore_pixiedust"] = "true"
            ipythonDisplay(Javascript(self.renderTemplate( "addScriptCode.js", type="css", code = self.renderTemplate("pixiedust.css") )))
            js = self.renderTemplate( "addScriptCode.js", type="javascript", code = self.renderTemplate("pixiedust.js") )
            #self.debug("pixiedust code: {}".format(js))
            ipythonDisplay(Javascript(js))
    
    def render(self):
        self._checkPixieDustJS()
        handlerId=self.options.get("handlerId")
        if handlerId is None or not self.noChrome:
            #get the first menuInfo for this handler and generate a js call
            menuInfo = globalMenuInfos[handlerId] if handlerId is not None and handlerId in globalMenuInfos else None
            if menuInfo is None:
                menuInfos = self.handlerMetadata.getMenuInfo(self.entity, self.dataHandler)
                if len(menuInfos)>0:
                    menuInfo = menuInfos[0]
            if menuInfo is not None:
                self._addHTML("""
                    <script>
                    ({0})();
                    </script>
                """.format(self._getExecutePythonDisplayScript(menuInfo)))
        else:
            start = time.clock()
            self.doRender(handlerId)
            self.executionTime = time.clock() - start

        #generate final HTML
        ipythonDisplay(HTML(self._wrapBeforeHtml() + self.html + self._wrapAfterHtml()))
        self._addScriptElements()
         
    @abstractmethod
    def doRender(self, handlerId):
        raise Exception("doRender method not implemented")

    def _addHTML(self, fragment):
        self.html+=fragment

    def _addHTMLTemplate(self, templateName, **kwargs):
        self._addHTML(self.renderTemplate(templateName, **kwargs))

    def renderTemplateString(self, source, **kwargs):
        return self.env.from_string(source).render(self._getTemplateArgs(**kwargs))

    def _addHTMLTemplateString(self, source, **kwargs):
        self._addHTML(
            self.renderTemplateString(source, **kwargs)
        )

    def _addJavascript(self, javascript):
        ipythonDisplay( Javascript(javascript) )

    def _addJavascriptTemplate(self, templateName, **kwargs):
        self._addJavascript(self.renderTemplate(templateName, **kwargs))
    
    def _safeString(self, s):
        if not isinstance(s, str if sys.version >= '3' else basestring):
            return str(s)
        else:
            return s.encode('ascii', 'ignore')
        
    #def _addD3Script2(self):
    #    print("Adding d3")
    #    self._addScriptElement("//cdnjs.cloudflare.com/ajax/libs/d3/3.4.8/d3.min")
        
    def _addScriptElement(self, script,checkJSVar=None, callback=None):
        self.scripts.append((script,checkJSVar,callback))
        
    def _addScriptElements(self):
        if len(self.scripts)==0:
            return
        ipythonDisplay(Javascript(self.renderTemplate('addScriptElements.js')))
    
    def _wrapBeforeHtml(self):
        if self.noChrome:
            return ""   
        menuTree=OrderedDict()
        for catId in ActionCategories.CAT_INFOS.keys():
            menuTree[catId]=[]
        for handler in (handlers+systemHandlers):
            for menuInfo in handler.getMenuInfo(self.entity, self.dataHandler):
                #fix the icon-path if available
                if "icon-path" in menuInfo and ":" not in menuInfo["icon-path"]:
                    menuInfo["icon-path"] = handler.__module__ + ":" + menuInfo["icon-path"]
                categoryId=menuInfo.get('categoryId')
                if categoryId is not None:
                    if not categoryId in menuTree:
                        menuTree[categoryId]=[menuInfo]
                    else:
                        menuTree[categoryId].append(menuInfo)
        return self.renderTemplate('cellOutput.html', menuTree=menuTree, numMenu=reduce(lambda n,t:n+t, [len(v) for k,v in iteritems(menuTree)], 0))
    
    def getPrefix(self, menuInfo=None):
        if ( not hasattr(self, 'prefix') ):
            self.prefix = self.options.get("prefix")
            if (self.prefix is None):
                self.prefix = str(uuid.uuid4())[:8]
        return self.prefix if menuInfo is None else (self.prefix + "-" + menuInfo['id'])
    
    def _getExecutePythonDisplayScript(self, menuInfo=None):
        return self.renderTemplate('executePythonDisplayScript.js',menuInfo=menuInfo)
        
    def _getMenuHandlerScript(self, menuInfo):
        return """
            <script>
                $('#menu{0}').on('click', {1})
            </script>
        """.format(self.getPrefix(menuInfo),self._getExecutePythonDisplayScript(menuInfo))
        
    def _genDisplayScript(self, menuInfo=None,addOptionDict={}):
        def updateCommand(command,key,value):
            hasValue = value is not None and value != ''
            replaceValue = (key+"=" + str(value)) if hasValue else ""
            pattern = ("" if hasValue else ",")+"\\s*" + key + "\\s*=\\s*'((\\\\'|[^'])*)'"
            m = re.search(pattern, str(command), re.IGNORECASE)
            retCommand = command
            if m is not None:
                retCommand = command.replace(m.group(0), key+"='"+value+"'" if hasValue else "");
            elif hasValue:
                k=command.rfind(")")
                retCommand = command[:k]
                retCommand+=","+key+"='"+value+"'"
                retCommand+= command[k:]
            return retCommand

        command = self.callerText
        if command is None:
            raise ValueError("command is None")
        if menuInfo:
            command = updateCommand(command, "handlerId", menuInfo['id'])
            command = updateCommand(command, "prefix", self.getPrefix())
        if "cell_id" not in self.options:
            command = updateCommand(command, "cell_id", 'cellId')
        for key,value in iteritems(addOptionDict):
            command = updateCommand(command, key, value)

        command = updateCommand(command, "showchrome", None)

        if "nostore_pixiedust" in self.options:
            command = updateCommand(command, "nostore_pixiedust", self.options["nostore_pixiedust"])
        #remove showchrome if there
        return command.replace("\"","\\\"")

    def getCategory(self, catId):
        if catId in ActionCategories.CAT_INFOS:
            return ActionCategories.CAT_INFOS[catId]
        myLogger.error("catId {0} not found in Registry".format(catId))
        
    def getCategoryTitle(self,catId):
        if catId in ActionCategories.CAT_INFOS:
            return ActionCategories.CAT_INFOS[catId]['title']
        elif catId == "Download":
            return "Stash dataset"
        myLogger.error("catId {0} not found in Registry".format(catId))
        return ""
            
    def getCategoryIconClass(self,catId):
        if catId in ActionCategories.CAT_INFOS:
            return ActionCategories.CAT_INFOS[catId]['icon-class']
        elif catId == "Download":
            return "fa-cloud-download"
        myLogger.error("catId {0} not found in Registry".format(catId))
        return ""
        
    def _wrapAfterHtml(self):
        if ( self.noChrome ):
            return ("""<div class="executionTime" id="execution{0}">Execution time: {1}s</div>""".format(self.getPrefix(), str(self.executionTime))) if self.executionTime is not None and self.addProfilingTime else ""
        return "</div>"

#Special handler for fetching the id of the cell being executed 
class CellHandshakeMeta(DisplayHandlerMeta):
    def getMenuInfo(self,entity, dataHandler):
       return []
    def newDisplayHandler(self,options,entity):
        return CellHandshake(options,entity)

@Logger()    
class CellHandshake(Display):
    snifferCallbacks = []

    def __init__(self, options, entity):
        Display.__init__(self, options, entity)
        self.nostore_params = True

    @staticmethod
    def addCallbackSniffer(sniffer):
        CellHandshake.snifferCallbacks.append(sniffer)
    def render(self):
        self._checkPixieDustJS()
        ipythonDisplay(HTML(
            self.renderTemplate("handshake.html", org_params = ','.join(list(self.options.keys())))
        ))
        
    def doRender(self, handlerId):
        pass

#Special handler for running in a dialog
class RunInDialogMeta(DisplayHandlerMeta):
    def getMenuInfo(self, entity, dataHandler):
        return []
    def newDisplayHandler(self, options, entity):
        return RunInDialog(options, entity)

@Logger()
class RunInDialog(Display):
    def render(self):
        self._checkPixieDustJS()
        self.debug("In RunInDialog")
        # del self.options['runInDialog']
        ipythonDisplay(Javascript("pixiedust.executeInDialog({0});".format(
            json.dumps({
                "prefix": self.getPrefix(),
                "command": self.callerText.replace(",runInDialog='true'",""),
                "options": self.options
            })
        )))
    def doRender(self, handlerId):
        pass

#Special handler used when no handlers was found to process the entity 
class UnknownEntityMeta(DisplayHandlerMeta):
    def getMenuInfo(self,entity, dataHandler):
       return []
    def newDisplayHandler(self,options,entity):
        return UnknownEntityDisplay(options,entity)
        
class UnknownEntityDisplay(Display):
    def render(self):
        ipythonDisplay(HTML(
            self.renderTemplate("unknownEntity.html")
        ))
        
    def doRender(self, handlerId):
        pass



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from six import iteritems
import pixiedust
from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.utils.template import PixiedustTemplateEnvironment
from pixiedust.utils.environment import Environment,scalaGateway
import uuid
import tempfile
from collections import OrderedDict
from IPython.display import display, HTML, Javascript
try:
    from urllib.request import Request, urlopen, URLError, HTTPError
except ImportError:
    from urllib2 import Request, urlopen, URLError, HTTPError

dataDefs = OrderedDict([
    ("1", {
        "displayName": "Car performance data", 
        "url": "https://github.com/ibm-cds-labs/open-data/raw/master/cars/cars.csv",
        "topic": "transportation",
        "publisher": "IBM",
        "schema2": [('mpg','int'),('cylinders','int'),('engine','double'),('horsepower','int'),('weight','int'),
            ('acceleration','double'),('year','int'),('origin','string'),('name','string')]
    }),
    ("2", {
        "displayName": "Sample retail sales transactions, January 2009", 
        "url": "https://raw.githubusercontent.com/ibm-cds-labs/open-data/master/salesjan2009/salesjan2009.csv",
        "topic": "Economy & Business",
        "publisher": "IBM Cloud Data Services"
    }),
    ("3", {
        "displayName": "Total population by country", 
        "url": "https://apsportal.ibm.com/exchange-api/v1/entries/889ca053a19986a4445839358a91963e/data?accessKey=657b130d504ab539947e51b50f0e338e",
        "topic": "Society",
        "publisher": "IBM Cloud Data Services"
    }),
    ("4", {
        "displayName": "GoSales Transactions for Naive Bayes Model", 
        "url": "https://apsportal.ibm.com/exchange-api/v1/entries/8044492073eb964f46597b4be06ff5ea/data?accessKey=bec2ed69d9c84bed53826348cdc5690b",
        "topic": "Leisure",
        "publisher": "IBM"
    }),
    ("5", {
        "displayName": "Election results by County", 
        "url": "https://openobjectstore.mybluemix.net/Election/county_election_results.csv",
        "topic": "Society",
        "publisher": "IBM"
    }),
    ("6", {
        "displayName": "Million dollar home sales in NE Mass late 2016", 
        "url": "https://openobjectstore.mybluemix.net/misc/milliondollarhomes.csv",
        "topic": "Economy & Business",
        "publisher": "Redfin.com"
    }),
    ("7", {
        "displayName": "Boston Crime data, 2-week sample", 
        "url": "https://raw.githubusercontent.com/ibm-cds-labs/open-data/master/crime/boston_crime_sample.csv",
        "topic": "Society",
        "publisher": "City of Boston"
    })
])

@scalaGateway
def sampleData(dataId=None):
    global dataDefs
    return SampleData(dataDefs).sampleData(dataId)

class SampleData(object):
    env = PixiedustTemplateEnvironment()
    def __init__(self, dataDefs):
        self.dataDefs = dataDefs

    def sampleData(self, dataId = None):
        if dataId is None:
            self.printSampleDataList()
        elif str(dataId) in dataDefs:
            return self.loadSparkDataFrameFromSampleData(dataDefs[str(dataId)])
        elif "https://" in str(dataId) or "http://" in str(dataId) or "file://" in str(dataId):
            return self.loadSparkDataFrameFromUrl(str(dataId))
        else:
            print("Unknown sample data identifier. Please choose an id from the list below")
            self.printSampleDataList()

    def printSampleDataList(self):
        display( HTML( self.env.getTemplate("sampleData.html").render( dataDefs = iteritems(self.dataDefs) ) ))
        #for key, val in iteritems(self.dataDefs):
        #    print("{0}: {1}".format(key, val["displayName"]))

    def dataLoader(self, path, schema=None):
        #TODO: if in Spark 2.0 or higher, use new API to load CSV
        load = ShellAccess["sqlContext"].read.format('com.databricks.spark.csv')
        if schema is not None:
            def getType(t):
                if t == 'int':
                    return IntegerType()
                elif t == 'double':
                    return DoubleType()
                else:
                    return StringType()
            return load.options(header='true', mode="DROPMALFORMED").load(path, schema=StructType([StructField(item[0], getType(item[1]), True) for item in schema]))
        else:
            return load.options(header='true', mode="DROPMALFORMED", inferschema='true').load(path)

    def loadSparkDataFrameFromSampleData(self, dataDef):
        return Downloader(dataDef).download(self.dataLoader)

    def loadSparkDataFrameFromUrl(self, dataUrl):
        i = dataUrl.rfind('/')
        dataName = dataUrl[(i+1):]
        dataDef = {
            "displayName": dataUrl,
            "url": dataUrl
        }
        return Downloader(dataDef).download(self.dataLoader)


class Downloader(object):
    def __init__(self, dataDef):
        self.dataDef = dataDef
        self.headers = {"User-Agent": "PixieDust Sample Data Downloader/1.0"}
        self.prefix = str(uuid.uuid4())[:8]
    
    def download(self, dataLoader):
        displayName = self.dataDef["displayName"]
        if "path" in self.dataDef:
            path = self.dataDef["path"]
        else:
            url = self.dataDef["url"]
            req = Request(url, None, self.headers)
            print("Downloading '{0}' from {1}".format(displayName, url))
            with tempfile.NamedTemporaryFile(delete=False) as f:
                self.write(urlopen(req), f)
                path = f.name
                self.dataDef["path"] = path = f.name
        if path:
            try:
                print("Creating pySpark DataFrame for '{0}'. Please wait...".format(displayName))
                return dataLoader(path, self.dataDef.get("schema", None))
            finally:
                print("Successfully created pySpark DataFrame for '{0}'".format(displayName))
            
    def report(self, bytes_so_far, chunk_size, total_size):
        if bytes_so_far == 0:
            display( HTML( """
                <div>
                    <span id="pm_label{0}">Starting download...</span>
                    <progress id="pm_progress{0}" max="100" value="0" style="width:200px"></progress>
                </div>""".format(self.prefix)
                )
            )
        else:
            percent = float(bytes_so_far) / total_size
            percent = round(percent*100, 2)
            display(
                Javascript("""
                    $("#pm_label{prefix}").text("{label}");
                    $("#pm_progress{prefix}").attr("value", {percent});
                """.format(prefix=self.prefix, label="Downloaded {0} of {1} bytes".format(bytes_so_far, total_size), percent=percent))
            )

    def write(self, response, file, chunk_size=8192):
        total_size = response.headers['Content-Length'].strip() if 'Content-Length' in response.headers else 100
        total_size = int(total_size)
        bytes_so_far = 0

        self.report(bytes_so_far, chunk_size, total_size)

        while 1:
            chunk = response.read(chunk_size)
            bytes_so_far += len(chunk)
            if not chunk:
                break
            file.write(chunk)             
            total_size = bytes_so_far if bytes_so_far > total_size else total_size
            self.report(bytes_so_far, chunk_size, total_size)

        return bytes_so_far


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from .storage import *

USER_PREFERENCES_TBL_NAME="USER_PREFERENCES"

class __UserPreferences(Storage):
    def __init__(self):
        self._initTable( USER_PREFERENCES_TBL_NAME,
        '''
            PREF_KEY     TEXT  NOT NULL,
            PREF_VALUE   TEXT  NOT NULL,
            PRIMARY KEY (PREF_KEY)
        ''')

userPrefStorage = __UserPreferences()

def getUserPreference(prefKey, defaultValue=None):
    retValue = userPrefStorage.fetchOne("""
            SELECT * from {0} WHERE PREF_KEY='{1}'
        """.format(USER_PREFERENCES_TBL_NAME,prefKey), 
        lambda row: row["PREF_VALUE"]
    )
    return defaultValue if retValue is None else retValue

def setUserPreference(prefKey, prefValue):
    if getUserPreference(prefKey) is not None:
        userPrefStorage.update("""
            UPDATE {0} SET PREF_VALUE='{2}' WHERE PREF_KEY = '{1}'
        """.format(USER_PREFERENCES_TBL_NAME,prefKey, prefValue))
    else:
        userPrefStorage.insert("""
            INSERT INTO {0} (PREF_KEY, PREF_VALUE)
            VALUES ('{1}','{2}')
        """.format(USER_PREFERENCES_TBL_NAME, prefKey,prefValue))



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.display import display
from pixiedust.display.display import *
from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.utils import Logger
from six import iteritems
from abc import ABCMeta
import inspect
import sys
from six import string_types

def route(**kw):
    def route_dec(fn):
        fn.pixiedust_route=kw
        return fn
    return route_dec

@Logger()
class PixieDustApp(Display):

    routesByClass = {}
    
    def matchRoute(self, route):
        for key,value in iteritems(route):
            #first check if the key is an field of the class
            option = getattr(self.entity, key) if self.entity is not None and hasattr(self.entity, key) else None
            #make sure we don't have a conflict with an existing function
            if callable(option):
                option = None
            if option is None:
                option = self.options.get(key,None)
            if  (option is None and value=="*") or (value != "*" and option != value):
                return False
        return True
        
    def doRender(self, handlerId):
        if self.__class__.__name__ in PixieDustApp.routesByClass:
            defRoute = None
            retValue = None
            try:
                dispatchKey = "widgets" if "widget" in self.options else "routes"
                for t in PixieDustApp.routesByClass[self.__class__.__name__][dispatchKey]:
                    if not t[0]:
                        defRoute = t[1]
                    elif self.matchRoute(t[0]):
                        self.debug("match found: {}".format(t[0]))
                        retValue = getattr(self, t[1])()
                        return
                if defRoute:
                    retValue = getattr(self, defRoute)()
                    return
            finally:
                if isinstance(retValue, string_types):
                    self._addHTMLTemplateString(retValue)
                elif isinstance(retValue, dict):
                    body = self.renderTemplateString(retValue.get("body", ""))
                    jsOnLoad = self.renderTemplateString(retValue.get("jsOnLoad", ""))
                    jsOK = self.renderTemplateString(retValue.get("jsOK", ""))
                    dialogRoot = retValue.get("dialogRoot", None)
                    if dialogRoot is not None:
                        jsOnLoad = """pixiedust.dialogRoot="{}";\n{}""".format(self.renderTemplateString(dialogRoot), jsOnLoad)
                    if body is not None:
                        self._addHTMLTemplateString("""
                        {{body}}
                        <pd_dialog>
                            <pd_onload>{{jsOnLoad|htmlAttribute}}</pd_onload>
                            <pd_ok>{{jsOK|htmlAttribute}}</pd_ok>
                        </pd_dialog>
                        """, body=body, jsOnLoad=jsOnLoad, jsOK=jsOK)

        print("Didn't find any routes for {}".format(self))

    def getDialogOptions(self):
        return {}

@Logger()
def PixieApp(cls):
    #reset the class routing in case the cell is being run multiple time
    clsName = "{}_{}_Display".format(inspect.getmodule(cls).__name__, cls.__name__)
    PixieDustApp.routesByClass[clsName] = {"routes":[], "widgets":[]}
    #put the routes that define a widget in a separate bucket

    def walk(cl):
        for name, method in iteritems(cl.__dict__):
            if hasattr(method, "pixiedust_route"):
                if "widget" in method.pixiedust_route:
                    PixieDustApp.routesByClass[clsName]["widgets"].append( (method.pixiedust_route,name) )
                else:
                    PixieDustApp.routesByClass[clsName]["routes"].append( (method.pixiedust_route,name) )
        for c in [c for c in cl.__bases__]:
            walk(c)
    walk(cls)

    #re-order the routes according to the number of constraints e.g. from more to less specific
    p = PixieDustApp.routesByClass[clsName]["routes"]
    PixieDustApp.routesByClass[clsName]["routes"] = [p[a[1]] for a in sorted([(len(a[0]), i) for i,a in enumerate(p)], reverse=True)]

    def __init__(self, options=None, entity=None, dataHandler=None):
        PixieDustApp.__init__(self, options or {}, entity, dataHandler)
        if hasattr(cls, "setup"):
            cls.setup(self)
        self.nostore_params = True

    def getPixieAppEntity(self):
        return self.pixieapp_entity if hasattr(self, "pixieapp_entity") else None

    def decoName(cls, suffix):
        return "{}_{}_{}".format(cls.__module__, cls.__name__, suffix)

    def run(self, entity=None, **kwargs):
        if entity is not None:
            self.pixieapp_entity = entity
        var = None
        for key in ShellAccess:
            if ShellAccess[key] is self:
                var = key

        if not var:
            #If we're here, the user must have created the instance inline, assign a variable dynamically
            var = cls.__name__ + "_instance"
            ShellAccess[var] = self

        self.runInDialog = kwargs.get("runInDialog", "false") is "true"
        options = {"runInDialog": "true" if self.runInDialog else "false"}
        if self.runInDialog:
            options.update( self.getDialogOptions() )

        options.update( {'handlerId': decoName(cls, "id") })

        s = "display({}{})".format(var, reduce(lambda k,v: k + "," + v[0] + "='" + v[1] + "'", iteritems(options), ""))
        try:
            sys.modules['pixiedust.display'].pixiedust_display_callerText = s
            locals()[var] = self
            return eval(s, globals(), locals())
        finally:
            del sys.modules['pixiedust.display'].pixiedust_display_callerText
        
    displayClass = type( decoName(cls, "Display"), (cls,PixieDustApp, ),{"__init__": __init__, "run": run, "getPixieAppEntity":getPixieAppEntity})
    
    @addId
    def getMenuInfo(self, entity, dataHandler=None):
        if entity is displayClass or entity.__class__ is displayClass:
            return [{"id": decoName(cls, "id")}]
        return []

    def newDisplayHandler(self, options, entity):
        entity.__init__(options, entity)
        return entity
    
    displayHandlerMetaClass = type( decoName(cls, "Meta"), (DisplayHandlerMeta,), {
            "getMenuInfo": getMenuInfo,
            "newDisplayHandler": newDisplayHandler
        })
    
    displayHandlerMeta = displayHandlerMetaClass()
    registerDisplayHandler( displayHandlerMeta )
    return displayClass


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display import *
from pixiedust.utils.dataFrameAdapter import *
from pixiedust.display.chart.renderers import PixiedustRenderer
import pixiedust

myLogger = pixiedust.getLogger(__name__ )

#bootstrap all the renderers
renderers = ["matplotlib", "bokeh", "seaborn", "google", "mapbox"]

for renderer in renderers:
    try:
        __import__("pixiedust.display.chart.renderers." + renderer)
    except ImportError as e:
        myLogger.warn("Unable to import renderer {0}: {1}".format(renderer, str(e)))

@PixiedustDisplayMeta()
class ChartDisplayMeta(DisplayHandlerMeta):
    @addId
    def getMenuInfo(self, entity, dataHandler):
        if dataHandler is not None:
            return [
                {"categoryId": "Chart", "title": "Bar Chart", "icon": "fa-bar-chart", "id": "barChart"},
                {"categoryId": "Chart", "title": "Line Chart", "icon": "fa-line-chart", "id": "lineChart"},
                {"categoryId": "Chart", "title": "Scatter Plot", "icon": "fa-circle", "id": "scatterPlot"},
                {"categoryId": "Chart", "title": "Pie Chart", "icon": "fa-pie-chart", "id": "pieChart"},
                {"categoryId": "Chart", "title": "Map", "icon": "fa-globe", "id": "mapView"},
                {"categoryId": "Chart", "title": "Histogram", "icon": "fa-table", "id": "histogram"}
            ]
        return []

    def newDisplayHandler(self, options, entity):
        return PixiedustRenderer.getRenderer(options, entity)


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
import pixiedust.utils.dataFrameMisc as dataFrameMisc
from pyspark.sql import functions as F
from pyspark.sql.types import DecimalType
import time
import pandas as pd
from pixiedust.utils import Logger

@Logger()
class PySparkDataFrameDataHandler(object):
    def __init__(self, options, entity):
        self.options = options
        self.entity = entity

    def __getattr__(self, name):
        if hasattr(self.entity, name):
            return self.entity.__getattribute__(name)
        raise AttributeError("{0} attribute not found".format(name))

    def getFieldNames(self, expandNested=False):
        return dataFrameMisc.getFieldNames(self.entity, expandNested)

    def isNumericField(self, fieldName):
        return dataFrameMisc.isNumericField(self.entity, fieldName)

    def isNumericType(self, field):
        return dataFrameMisc.isNumericType(field.dataType)

    def isStringField(self, fieldName):
        return dataFrameMisc.isStringField(self.entity, fieldName)

    def isStringType(self, field):
        return dataFrameMisc.isStringType(field.dataType)

    def isDateField(self, fieldName):
        return dataFrameMisc.isDateField(self.entity, fieldName)

    def isDateType(self, field):
        return dataFrameMisc.isDateType(field.dataType)

    def getFieldValues(self, fieldNames):
        if len(fieldNames) == 0:
            return []
        numericKeyField = False
        if len(keyFields) == 1 and self.isNumericField(fieldNames[0]):
            numericKeyField = True
        df = self.entity.groupBy(fieldNames).count().dropna()
        for fieldName in fieldNames:
            df = df.sort(F.col(fieldName).asc())
        maxRows = int(self.options.get("rowCount","100"))
        numRows = min(maxRows,df.count())
        rows = df.take(numRows)
        values = []
        for i, row in enumerate(rows):
            if numericKeyField:
                values.append(row[fieldNames[0]])
            else:
                values.append(i)
        return values

    """
        Return a cleaned up Pandas Dataframe that will be used as working input to the chart
    """
    def getWorkingPandasDataFrame(self, xFields, yFields, extraFields=[], aggregation=None, maxRows = 100):
        if xFields is None or len(xFields)==0:
            #swap the yFields with xFields
            xFields = yFields
            yFields = []
            aggregation = None

        extraFields = [a for a in extraFields if a not in xFields]
        workingDF = self.entity.select(xFields + extraFields + yFields)
        if aggregation and len(yFields)>0:
            aggMapper = {"SUM":"sum", "AVG": "avg", "MIN": "min", "MAX": "max"}
            aggregation = aggMapper.get(aggregation, "count")

            workingDF = workingDF.groupBy(extraFields + xFields).agg(dict([(yField, aggregation) for yField in yFields]))            
            for yField in yFields:
                workingDF = workingDF.withColumnRenamed("{0}({1})".format(aggregation,yField), yField)

        workingDF = workingDF.dropna()
        count = workingDF.count()
        if count > maxRows:
            workingDF = workingDF.sample(False, (float(maxRows) / float(count)))
        pdf = self.toPandas(workingDF)

        #check if the user wants timeseries
        if len(xFields) == 1 and self.options.get("timeseries", 'false') == 'true':
            field = xFields[0]
            try:
                pdf[field] = pd.to_datetime(pdf[field])
            except:
                self.exception("Unable to convert field {} to datetime".format(field))

        #sort by xFields
        pdf.sort_values(xFields + extraFields, inplace=True)
        return pdf

    """
    Custom implementation of toPandas. It checks the spark type of each column in the dataframe for DecimalType. If any are found, it check the 
    corresponding pandas dataframe column to make sure it's not a python object type (which would cause issue during plotting). If that's the case, 
    it cast them as float
    """
    def toPandas(self, workingDF):        
        decimals = []
        for f in workingDF.schema.fields:
            if f.dataType.__class__ == DecimalType:
                decimals.append(f.name)

        pdf = workingDF.toPandas()
        for y in pdf.columns:
            if pdf[y].dtype.name == "object" and y in decimals:
                #spark converts Decimal type to object during toPandas, cast it as float
                pdf[y] = pdf[y].astype(float)

        return pdf




# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.utils import Logger
from pixiedust.utils.storage import *
import json

CONNECTION_TBL_NAME = "service_connections"
VERSION = 1 

#global Storage
@Logger()
class __ServiceManagerStorage(Storage):
    def __init__(self):
        self._initTable( CONNECTION_TBL_NAME, 
        """
            NAME            TEXT NOT NULL,
            TYPE            TEXT NOT NULL,
            VERSION         INTEGER NOT NULL,
            PAYLOAD         TEXT  NOT NULL,
            PRIMARY KEY (NAME, TYPE)
        """)

    def getConnections(self, connectionType):
        return self.fetchMany("""
                SELECT * FROM {0} WHERE TYPE = '{1}'
            """.format(CONNECTION_TBL_NAME, connectionType),
            lambda row: json.loads(row["PAYLOAD"])
        )

    def getConnection(self, connectionType, connectionName, raw=True):
        row = self.fetchOne("""
            SELECT * FROM {0} WHERE NAME='{1}' AND TYPE='{2}'
        """.format(
                CONNECTION_TBL_NAME,
                connectionName,
                connectionType
            )
        )
        if row is None:
            raise Exception("Unable to find connection {0} with type {1}".format(connectionName, connectionType))

        return row if raw else json.loads(row['PAYLOAD'])

    def addConnection(self, connectionType, payload):
        self._validatePayload(payload)

        name = payload["name"]
        try:
            self.getConnection(connectionType, name )
            #already exist, update
            self.update("""
                UPDATE {0} SET PAYLOAD='{1}' WHERE NAME = '{2}' and TYPE='{3}'
            """.format(CONNECTION_TBL_NAME, json.dumps(payload), payload["name"], connectionType))
        except:
            #doesn't exist, try to insert
            self.insert("""
                INSERT INTO {0} (NAME, TYPE,VERSION,PAYLOAD)
                VALUES ('{1}','{2}','{3}','{4}')
            """.format(CONNECTION_TBL_NAME, payload["name"], connectionType, VERSION, json.dumps(payload)))

        return payload["name"]

    def _validatePayload(self, payload):
        if not "name" in payload:
            raise Exception("Missing field name")

    def deleteConnection(self, connectionType, connectionName):
        return self.delete("""
            DELETE FROM {0} WHERE NAME='{1}' AND TYPE='{2}'
        """.format(
                CONNECTION_TBL_NAME,
                connectionName,
                connectionType
            )
        )

__connectionsStorage = __ServiceManagerStorage()

#public CRUD APIs
def getConnections(connectionType):
    return __connectionsStorage.getConnections(connectionType)

def getConnection(connectionType, connectionName, raw=True):
    return __connectionsStorage.getConnection(connectionType, connectionName, raw=raw)

def addConnection(connectionType, payload):
    return __connectionsStorage.addConnection(connectionType, payload)

def deleteConnection(connectionType, connectionName):
    return __connectionsStorage.deleteConnection(connectionType, connectionName)


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.utils import *
from pixiedust.utils.environment import Environment
from six import PY2

"""
return True is entity is a spark DataFrame
"""
def isPySparkDataFrame(entity):
    return fqName(entity)=="pyspark.sql.dataframe.DataFrame"

"""
return True is entity is a Pandas DataFrame
"""
def isPandasDataFrame(entity):
    return fqName(entity)=="pandas.core.frame.DataFrame"

def checkIfDataType(type):
    if Environment.hasSpark:
        from pyspark.sql.types import DataType
        return isinstance(type, DataType)
    else:
        return fqName(type).endswith("Type")

"""
Return a list of field names for the given dataframe
will expand nested structures if expandNested is set to True
"""
def getFieldNames(df, expandNested=False):
    def getFieldName(field, expand):
        if fqName(field.dataType) == "pyspark.sql.types.StructType" and expand:
            retFields = []
            for f in field.dataType.fields:
                retFields.extend([field.name + "." + name for name in getFieldName(f, expand)])
            return retFields
        else:
            return [field.name.decode('utf-8') if PY2 else field.name]
    fieldNames = []
    for field in df.schema.fields:
        fieldNames.extend( getFieldName(field, expandNested) )
    return fieldNames

"""
Return True if spark type is numeric
"""
def isNumericType(type):
    if checkIfDataType(type):
        return isNumericType( type.__class__.__name__)
    return (type =="LongType" or type == "IntegerType" or type == "DoubleType" or type == "DecimalType" or type == "FloatType")

"""
Return True if spark type is a string
"""
def isStringType(type):
    if checkIfDataType(type):
        return isStringType( type.__class__.__name__)
    return (type =="StringType")

"""
Return True if spark type is a date
"""
def isDateType(type):
    if checkIfDataType(type):
        return isDateType( type.__class__.__name__)
    return (type =="DateType")

"""
Return True is field represented by fieldName is Numeric
"""
def isNumericField(entity, fieldName):
    def isNumericFieldRecurse(field, targetName):
        if field.name == targetName:
            return isNumericType(field.dataType)
        elif fqName(field.dataType)  == "pyspark.sql.types.StructType" and targetName.startswith(field.name + "."):
            nestedFieldName = targetName[len(field.name)+1:]
            for f in field.dataType.fields:
                if isNumericFieldRecurse(f, nestedFieldName):
                    return True         
        return False
    for field in entity.schema.fields:
        if isNumericFieldRecurse(field, fieldName):
            return True
    return False

"""
Return True is field represented by fieldName is a String
"""
def isStringField(entity, fieldName):
    def isStringFieldRecurse(field, targetName):
        if field.name == targetName:
            return isStringType(field.dataType)
        elif checkIfDataType(field.dataType) and targetName.startswith(field.name + "."):
            nestedFieldName = targetName[len(field.name)+1:]
            for f in field.dataType.fields:
                if isStringFieldRecurse(f, nestedFieldName):
                    return True         
        return False
    for field in entity.schema.fields:
        if isStringFieldRecurse(field, fieldName):
            return True
    return False

"""
Return True is field represented by fieldName is a Date
"""
def isDateField(entity, fieldName):
    def isDateFieldRecurse(field, targetName):
        if field.name == targetName:
            return isDateType(field.dataType)
        elif checkIfDataType(field.dataType) and targetName.startswith(field.name + "."):
            nestedFieldName = targetName[len(field.name)+1:]
            for f in field.dataType.fields:
                if isDateFieldRecurse(f, nestedFieldName):
                    return True         
        return False
    for field in entity.schema.fields:
        if isDateFieldRecurse(field, fieldName):
            return True
    return False


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display import *
from .test1 import *
from .test2 import *

#Make sure that matplotlib is running inline
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    #get_ipython().run_line_magic("matplotlib", "inline")

class TestsDisplayMeta(DisplayHandlerMeta):
    @addId
    def getMenuInfo(self,entity, dataHandler):
        if safeCompare(entity, "test1"):
            return [
                {"categoryId": "Test", "title": "Test1", "icon": "fa-bar-chart", "id": "Test1"}
            ]
        elif safeCompare(entity, "test2"):
            return [
                {"categoryId": "Test", "title": "Test2", "icon": "fa-bar-chart", "id": "Test2"}
            ]
        else:
            return []
    def newDisplayHandler(self,options,entity):
        handlerId=options.get("handlerId")
        if handlerId == "Test1":
            return Test1Display(options,entity)
        else:
            return Test2Display(options,entity)

registerDisplayHandler(TestsDisplayMeta())


# coding=utf-8
# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from .seabornBaseDisplay import SeabornBaseDisplay
from pixiedust.utils import Logger
import numpy as np
import pandas as pd
import seaborn as sns

@PixiedustRenderer(id="barChart")
@Logger()
class sbBarChartDisplay(SeabornBaseDisplay):
    def renderStacked(self, ax):
        pandaList = self.getWorkingPandasDataFrame()
        valueFieldValues = pandaList if len(pandaList) >= 1 else []
        key = self.getKeyFields()[0]
        colors = ['#79c36a','#f1595f','#599ad3','#f9a65a','#9e66ab','#cd7058','#d77fb3','#727272']

        #data = self.sumDfList(valueFieldValues, [åkey], "agg")

        for i, valueField in enumerate(self.getValueFields()):
            sns.barplot(x=key, y=valueField, data=self.getWorkingPandasDataFrame(), ax=ax, color=colors[i], orient=self.getOrientation() )
            #data = self.minusDf(data, valueFieldValues[i], [key], "agg")

        #sns.barplot(x=key, y='agg', data=self.getWorkingPandasDataFrame(), ax=ax, color=colors[0])

    def renderGrouped(self, ax):
        keyFields = self.getKeyFields()
        valueFields = self.getValueFields()

        df = self.getWorkingPandasDataFrame()
        cols = df.columns.values.tolist()
        selectedcols = np.concatenate([keyFields, valueFields])

        for col in cols:
            if col not in selectedcols:
                df.drop(col, axis=1, inplace=True)

        df1 = pd.melt(df, id_vars=keyFields).sort_values(['variable','value'])
        sns.barplot(x=keyFields[0], y='value', hue='variable', ax=ax, data=df1, orient=self.getOrientation())

    def matplotlibRender(self, fig, ax):
        data = None
        x = None
        y = None
        stacked = self.options.get("stacked", "false") == "true"
        if len(self.getValueFields())>1 and stacked:
            self.renderStacked(ax)
        elif len(self.getValueFields())>1:
            self.renderGrouped(ax)
        else:
            self.renderStacked(ax)

    def getOrientation(self):
        return "v" if self.options.get("orientation", "vertical") == "vertical" else "h"
  
    @commonChartOptions
    def getChartOptions(self):
        options = []
        if len(self.getValueFields()) > 1:
            options.insert(0,
                {
                'name': 'stacked',
                'description': 'Stacked Bar Chart',
                'metadata': {
                    'type': 'checkbox',
                    'default': "false"
                }
                }
            )
        return options

    def sumDfList(self, list, groupby, column):
        data = None
        for df in list:
            if data is None:
                data = df
            else:
                data = pd.concat([data, df]).groupby(groupby, as_index=False)[column].sum()
        return data

    def minusDf(self, minuend, subtrahend, groupby, column):
        return pd.concat([minuend, subtrahend]).groupby(groupby, as_index=False)[column].agg(np.subtract)


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Inherited from maven-artifact https://github.com/hamnis/maven-artifact
# -------------------------------------------------------------------------------
from traitlets.config.application import Application
from jupyter_client.kernelspecapp  import ListKernelSpecs
from traitlets import Unicode, Dict
from .createKernel import PixiedustInstall
from .generate import PixiedustGenerate

class PixiedustList(ListKernelSpecs):
    pass

class PixiedustJupyterApp(Application):
    subcommands = Dict({
        'install': (PixiedustInstall, "Install Kernels locally for development with Pixiedust"),
        'list': (PixiedustList, "List of pixiedust kernels"),
        'generate': (PixiedustGenerate, "Generate boiler plate code for a PixieDust plugin")
    })
    def start(self):
        if self.subapp is None:
            print('Invalid syntax: You need to specify a subcommand from this list {0}'.format(list(self.subcommands)))
            print()
            self.print_subcommands()
            self.exit(1)
        else:
            return self.subapp.start()

def main():
    PixiedustJupyterApp.launch_instance()


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display import Display
from functools import reduce
import time

DELIMITER="@#$DELIMITER@#$"

class DownloadFileHandler(Display):
    def doRender(self, handlerId):
        entity=self.entity
        self.addProfilingTime = False
        total=entity.count()
        schema = entity.schema
        doDownload=self.options.get("doDownload")
        doDownloadCount=self.options.get("doDownloadCount")
        doDownloadLink=self.options.get("doDownloadLink")
        if total==0:
            self._addHTML("<p>No data available for download</p>")
        elif doDownload is None:    
            self._addHTMLTemplate("downloadFile.html", totalDocs=total)
        elif doDownloadLink is None:
            self._addHTML("""<script>
                var blobs = [];
                var callbacks = {{
                    iopub:{{
                        output:function(msg) {{
                            var msg_type = msg.header.msg_type;
                            var content = msg.content;
                            if (msg_type==='stream') {{
                                var s = content.text.split("\\n");
                                if (s[0]==="{1}") {{
                                    for (var i = 1; i < s.length; i++ ) {{
                                        if (s[i]==="{1}") {{
                                            console.log('Starting download');
                                            var URL = (window.webkitURL || window.URL);
                                            var url = URL.createObjectURL(new Blob(blobs, {{type : "text/plain"}}));
                                            var a = document.createElement('a');
                                            a.innerHTML = "{2}.{4}";
                                            a.href = url;
                                            a.download = "{2}.{4}";
                                            
                                            var d = document.createElement('div');
                                            d.style = 'padding: 20px';
                                            d.innerHTML = '<span>File successfully exported for download: </span>';
                                            d.appendChild(a);

                                            document.getElementById('results{3}').innerHTML = '';
                                            document.getElementById('results{3}').appendChild(d);

                                            a.click();
                                            //return URL.revokeObjectURL(url);
                                        }}else{{
                                            blobs.push(new Blob([s[i]+"\\n"], {{type : "text/plain"}}));
                                        }}
                                    }}
                                }}
                                else {{
                                    $('#wrapperHTML{3}').html('<p>' + s.join('<br/>') + '</p>');
                                }}
                            }}
                            else if (msg_type === 'error') {{
                                var d = document.getElementById('wrapperHTML{3}');
                                d.appendChild('<span>Export for download failed: ' + content.ename + '</span>');
                                for (m in msg) console.log('#msg', m, msg[m]);
                            }}
                        }}
                    }}
                }};

                IPython.notebook.session.kernel.execute(
                    "{0}", 
                    callbacks, {{silent:false,store_history:false,stop_on_error:true}}
                );
                
                console.log("running download command: {0}");
            </script>""".format(self._genDisplayScript(addOptionDict={'doDownloadLink':'True'}),DELIMITER,"dataframe-"+time.strftime('%Y%m%d-%H%M%S'),self.getPrefix(), doDownload)
            )
        elif doDownload == "csv":
            #write headers first
            print(DELIMITER)
            print(reduce(lambda s,f: s + ("," if s!="" else "") + f.name, schema.fields,""))                
            for row in entity.take(doDownloadCount):
                print(reduce(lambda s,f: s+("," if s!="" else "")+self._safeString(row[f.name]), schema.fields, ""))
            print(DELIMITER)
        elif doDownload == "json":
            print(DELIMITER)
            print("[")
            for count, row in enumerate(entity.take(doDownloadCount), start=1):
                print(" {")
                print(reduce(lambda s,f: s+(",\n  " if s!="" else "  ")+"\""+self._safeString(f.name)+"\":"+(str(row[f.name]) if isinstance(row[f.name],(int,long, float)) else self._safeString("\""+("" if row[f.name] is None else row[f.name])+"\"")), schema.fields, ""))
                print(" }," if count != doDownloadCount else " }")
            print("]")
            print(DELIMITER)
        elif doDownload == "xml":
            print(DELIMITER)
            print('<?xml version="1.0" encoding="UTF-8"?>')
            print("<dataframe>")
            for count, row in enumerate(entity.take(doDownloadCount), start=1):
                print(" <row>")
                print(reduce(lambda s,f: s+("\n  " if s!="" else "  ")+"<"+self._safeString(f.name.replace (" ", "_"))+">"+self._safeString(row[f.name])+"</"+self._safeString(f.name.replace (" ", "_"))+">", schema.fields, ""))
                print(" </row>")
            print("</dataframe>")
            print(DELIMITER)
        elif doDownload == "html":
            print(DELIMITER)
            print("<table>\n <thead>\n  <tr>")
            print(reduce(lambda s,f: s+("\n   " if s!="" else "   ")+"<th>"+self._safeString(f.name)+"</th>", schema.fields, ""))
            print("  </tr>\n </thead>")
            print(" <tbody>")
            for count, row in enumerate(entity.take(doDownloadCount), start=1):
                print("  <tr>")
                print(reduce(lambda s,f: s+("\n   " if s!="" else "   ")+"<td>"+self._safeString(row[f.name])+"</td>", schema.fields, ""))
                print("  </tr>")
            print(" </tbody>\n</table>")
            print(DELIMITER)
        elif doDownload == "md":
            print(DELIMITER)
            print(reduce(lambda s,f: s+self._safeString(f.name)+"|", schema.fields, "|"))
            print(reduce(lambda s,f: s+"---|", schema.fields, "|"))
            for count, row in enumerate(entity.take(doDownloadCount), start=1):
                print(reduce(lambda s,f: s+self._safeString(row[f.name])+"|", schema.fields, "|"))
            print(DELIMITER)
        elif doDownload == "txt":
            print(DELIMITER)
            print(entity.show(doDownloadCount))
            print(DELIMITER)
        else:
            print("Selected option not implemented")


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Inherited from maven-artifact https://github.com/hamnis/maven-artifact
# -------------------------------------------------------------------------------

from jupyter_core.application import JupyterApp
from jinja2 import Environment, PackageLoader
from collections import OrderedDict
from six.moves import input
from six import iteritems
import os

class PixiedustGenerate(JupyterApp):
    env = Environment(loader=PackageLoader('install', 'templates'))

    def __init__(self, **kwargs):
        super(PixiedustGenerate, self).__init__(**kwargs)
        self.step = 1

    def hilite(self, message):
        return '\x1b[%sm%s\x1b[0m' % (';'.join(['32', '1']), message)

    def confirm(self, message, action="Keep"):
        answer = input(self.hilite(message) + "\n\t" + action + " y/n [y]? ")
        return 'y' if answer == '' else answer

    def input(self, message, validate=None, allowEmpty=False):
        value = None
        while value == None:
            value = input(self.hilite(message) )
            if not allowEmpty and value == "":
                value=None
            if validate is not None:
                v = validate(value)
                if not v[0]:
                    print(v[1])
                    value = None
        return value

    def start(self):
        try:
            self.files = []
            self.doStart()
        except KeyboardInterrupt:
            print("\nPixieDust generator aborted by user")

    def getStep(self):
        step = self.step
        self.step += 1
        return step

    def doStart(self):
        self.projectName = self.input(self.hilite("Step {}: Please enter the name of your project: ".format( self.getStep() )))
        self.location = os.getcwd()
        answer = self.confirm(self.hilite("Step {0}: Directory for your project: {1}".format( self.getStep(), self.location )))
        if answer != 'y':
            self.location = self.input(
                self.hilite("Please enter a directory for your project: "),
                lambda _: (os.path.exists(_), "Directory {} does not exist. Please try again or CTRL+C to abort".format(_))
            )
        
        #Check if directory already exists
        self.fullPath = os.path.join(self.location, self.projectName )
        if os.path.exists( self.fullPath ):
            print("Folder {} already exist. Exiting...".format(self.fullPath))
            return

        self.projectAuthor = self.input(self.hilite("Step {}: [Optional] Please enter the project author: ".format(self.getStep() )), allowEmpty=True)
        self.projectAuthorEmail = self.input(self.hilite("Step {}: [Optional] Please enter the project author email address: ".format(self.getStep() )), allowEmpty=True)
        self.projectUrl = self.input(self.hilite("Step {}: [Optional] Please enter the project url: ".format(self.getStep() )), allowEmpty=True)

        projectTypes = OrderedDict([
            ("1", ("Display Visualization", self.doDisplayVisualization)),
            ("2", ("Chart Renderer", self.doChartRenderer))
        ])
        print(self.hilite("Step {}: Please enter the project type:".format( self.getStep())))
        for key, value in iteritems(projectTypes):
            print("\t{0}.{1}".format(key, value[0]))
        projectType = self.input(self.hilite("Project type:"))

        if projectType not in projectTypes:
            print("Project Type {0} is not supported. Exiting...".format(projectType) )

        #Create the directory
        os.mkdir(self.fullPath)
        os.mkdir(os.path.join(self.fullPath, self.projectName))

        #Create the project artifacts
        self.writeFile(["setup.pytpl","LICENSE.tpl","MANIFEST.intpl","readme.mdtpl","setup.cfgtpl"])

        #Call out the generator function specific to the project type
        projectTypes.get(projectType)[1]()

    def writeFile(self, templates, dirName = None, targetFile=None):
        fullPath = self.fullPath if dirName is None else os.path.join(self.fullPath, dirName)
        for template in templates:
            if template.endswith("tpl"):
                if targetFile is None or len(templates) > 1:
                    targetFile = template[:-3]
                    if targetFile.endswith("."):
                        targetFile = targetFile[:-1]
                    index = targetFile.rfind('/')
                    targetFile = targetFile[index+1:] if index >= 0 else targetFile
                fullPathTargetFile = os.path.join(fullPath, targetFile)
                self.files.append(fullPathTargetFile)
                with open( fullPathTargetFile, "wb" ) as f:
                    f.write(PixiedustGenerate.env.get_template(template).render(this=self).encode("utf-8","ignore"))


    def doDisplayVisualization(self):
        print("Generating Display Visualization bootstrap code...")
        self.displayClassName = self.input(self.hilite(
            "Step {0}. Please enter the Display Handler class name. e.g {1}Handler: ".format(self.getStep(), self.projectName.capitalize())
        ))
        self.writeFile(["display/__init__.pytpl", "display/display.pytpl"], dirName=self.projectName)

        templateDir = os.path.join(self.fullPath,self.projectName,"templates")
        os.mkdir(templateDir)
        self.writeFile(["display/helloWorld.htmltpl"], dirName=templateDir)

        self.displaySuccess()

    def doChartRenderer(self):
        print("Generating Chart Renderer bootstrap code")
        self.rendererClassName = self.input(self.hilite(
            "Step {0}. Please enter the prefix for renderer class name. e.g {1}: ".format(self.getStep(), self.projectName.capitalize())
        ))

        chartIds = [("barChart","bar"),("lineChart","line"),("scatterPlot","scatter"),
            ("pieChart","pie"),("mapView",None),("histogram","hist")]
        self.chartIds = self.input(self.hilite(
            "Step {0}. Please enter one or more chart id (comma separated) to be associated with this renderer. e.g {1}. Leave empty to add all: ".format(self.getStep(), [c[0] for c in chartIds] )
        ),allowEmpty=True)

        if self.chartIds == "":
            self.chartIds = [c[0] for c in chartIds]
        else:
            self.chartIds = self.chartIds.split(",")

        #write the init and base
        self.writeFile(["chart/__init__.pytpl", "chart/rendererBaseDisplay.pytpl"], dirName=self.projectName)

        for c in self.chartIds:
            self.chartId = c
            knownCharts = [x for x in chartIds if x[0]==c]
            self.plotKind = "line" if len(knownCharts)==0 or knownCharts[0][1] is None else knownCharts[0][1]
            self.writeFile(["chart/rendererDisplay.pytpl"], dirName=self.projectName, targetFile="{0}Display.py".format(c))

        self.displaySuccess()

    def displaySuccess(self):
        print("Successfully generated boostrap code with following files:")
        for f in self.files:
            print("\t{}".format(f))
